from collections import Counter
from functools import singledispatch
from typing import Any, Iterable, List, Optional, Set, Tuple, TypeVar, Union

import simplejson as json

from jubeatools import song as jbt
from jubeatools.formats.filetypes import SongFile
from jubeatools.utils import none_or

from ..tools import make_memon_dumper
from . import schema as memon


def _dump_memon_1_0_0(song: jbt.Song, **kwargs: Any) -> SongFile:
    metadata = dump_metadata(song.metadata)
    common_timing = dump_file_timing(song)
    charts = {
        diff: dump_chart(chart, common_timing) for diff, chart in song.charts.items()
    }
    file = memon.File(
        version="1.0.0",
        metadata=metadata,
        timing=common_timing,
        data=charts,
    )
    json_file = memon.FILE_SCHEMA.dump(file)
    file_bytes = json.dumps(json_file, indent=4, use_decimal=True).encode("utf-8")
    return SongFile(contents=file_bytes, song=song)


dump_memon_1_0_0 = make_memon_dumper(_dump_memon_1_0_0)


def dump_metadata(metadata: jbt.Metadata) -> memon.Metadata:
    return memon.Metadata(
        title=metadata.title,
        artist=metadata.artist,
        audio=none_or(str, metadata.audio),
        jacket=none_or(str, metadata.cover),
        preview=(
            none_or(str, metadata.preview_file)
            or none_or(dump_preview, metadata.preview)
        ),
    )


def dump_preview(preview: jbt.Preview) -> memon.PreviewSample:
    return memon.PreviewSample(
        start=preview.start,
        duration=preview.length,
    )


def dump_file_timing(song: jbt.Song) -> Optional[memon.Timing]:
    events = get_common_value(
        t.events for _, _, t in song.iter_charts_with_applicable_timing()
    )
    beat_zero_offset = get_common_value(
        t.beat_zero_offset for _, _, t in song.iter_charts_with_applicable_timing()
    )
    hakus = get_common_value(
        none_or(frozenset.__call__, c.hakus or song.common_hakus)
        for c in song.charts.values()
    )
    timing = dump_timing(
        events=events,
        beat_zero_offset=beat_zero_offset,
        hakus=hakus,
    )
    timing.remove_default_values()
    return timing or None


T = TypeVar("T")


def get_common_value(values: Iterable[T]) -> Optional[T]:
    possible_values = Counter(values)
    value, count = possible_values.most_common(1).pop()
    if count >= 2:
        return value
    else:
        return None


def dump_timing(
    events: Optional[Iterable[jbt.BPMEvent]],
    beat_zero_offset: Optional[jbt.SecondsTime],
    hakus: Optional[Set[jbt.BeatsTime]],
) -> memon.Timing:
    return memon.Timing(
        offset=beat_zero_offset,
        resolution=None,
        bpms=none_or(dump_bpms, events),
        hakus=none_or(dump_hakus, hakus),
    )


def dump_bpms(bpms: Iterable[jbt.BPMEvent]) -> List[memon.BPMEvent]:
    return [dump_bpm(b) for b in bpms]


def dump_bpm(bpm: jbt.BPMEvent) -> memon.BPMEvent:
    return memon.BPMEvent(beat=beats_to_best_form(bpm.time), bpm=bpm.BPM)


def beats_to_best_form(b: jbt.BeatsTime) -> memon.SymbolicTime:
    if is_expressible_as_240th(b):
        return (240 * b.numerator) // b.denominator
    else:
        return beat_to_fraction_tuple(b)


def is_expressible_as_240th(b: jbt.BeatsTime) -> bool:
    return (240 * b.numerator) % b.denominator == 0


def beat_to_fraction_tuple(b: jbt.BeatsTime) -> Tuple[int, int, int]:
    integer_part = int(b)
    remainder = b % 1
    return (
        integer_part,
        remainder.numerator,
        remainder.denominator,
    )


def dump_hakus(hakus: Set[jbt.BeatsTime]) -> List[memon.SymbolicTime]:
    return [beats_to_best_form(b) for b in sorted(hakus)]


def dump_chart(chart: jbt.Chart, file_timing: Optional[memon.Timing]) -> memon.Chart:
    return memon.Chart(
        level=chart.level,
        resolution=None,
        timing=dump_chart_timing(chart.timing, chart.hakus, file_timing),
        notes=[dump_note(n) for n in chart.notes],
    )


def dump_chart_timing(
    chart_timing: Optional[jbt.Timing],
    chart_hakus: Optional[Set[jbt.BeatsTime]],
    file_timing: Optional[memon.Timing],
) -> Optional[memon.Timing]:
    if chart_timing is None:
        events = None
        beat_zero_offset = None
    else:
        events = chart_timing.events
        beat_zero_offset = chart_timing.beat_zero_offset

    res = dump_timing(events, beat_zero_offset, chart_hakus)
    fallback = memon.Timing.fill_in_defaults(file_timing)
    return res.remove_common_values(fallback)


@singledispatch
def dump_note(n: Union[jbt.TapNote, jbt.LongNote]) -> memon.Note:
    raise ValueError(f"Unexpected argument : {n}")


@dump_note.register
def dump_tap_note(tap: jbt.TapNote) -> memon.TapNote:
    return memon.TapNote(n=tap.position.index, t=beats_to_best_form(tap.time))


@dump_note.register
def dump_long_note(long: jbt.LongNote) -> memon.LongNote:
    return memon.LongNote(
        n=long.position.index,
        t=beats_to_best_form(long.time),
        l=beats_to_best_form(long.duration),
        p=tail_as_6_notation(long),
    )


def tail_as_6_notation(long: jbt.LongNote) -> int:
    if long.tail_tip.y == long.position.y:
        return long.tail_tip.x - int(long.tail_tip.x > long.position.x)
    else:
        return 3 + long.tail_tip.y - int(long.tail_tip.y > long.position.y)



"""
memon (memo + json)
□□□■——◁

memon is a json-based jubeat chart set format designed to be easier to
parse than existing "memo-like" formats (memo, youbeat, etc ...).

https://github.com/Stepland/memon
"""

from .v0.dump import (
    dump_memon_0_1_0,
    dump_memon_0_2_0,
    dump_memon_0_3_0,
    dump_memon_legacy,
)
from .v0.load import (
    load_memon_0_1_0,
    load_memon_0_2_0,
    load_memon_0_3_0,
    load_memon_legacy,
)
from .v1.dump import dump_memon_1_0_0
from .v1.load import load_memon_1_0_0



from pathlib import Path
from typing import Any, Callable, Dict

import simplejson as json

from jubeatools import song as jbt
from jubeatools.formats.dump_tools import FileNameFormat
from jubeatools.formats.load_tools import FolderLoader, make_folder_loader
from jubeatools.formats.typing import Dumper, Loader, SongFileDumper


def _load_raw_memon(path: Path) -> Any:
    with path.open() as f:
        return json.load(f, use_decimal=True)


load_folder: FolderLoader[Any] = make_folder_loader("*.memon", _load_raw_memon)


def make_memon_folder_loader(memon_loader: Callable[[Any], jbt.Song]) -> Loader:
    """Create memon folder loader from the given file loader"""

    def load(path: Path, merge: bool = False, **kwargs: Any) -> jbt.Song:
        files = load_folder(path)
        if not merge and len(files) > 1:
            raise ValueError(
                "Multiple .memon files were found in the given folder, "
                "use the --merge option if you want to make a single memon file "
                "out of several that each containt a different chart (or set of "
                "charts) for the same song"
            )

        charts = [memon_loader(d) for d in files.values()]
        return jbt.Song.from_monochart_instances(*charts)

    return load


def make_memon_dumper(internal_dumper: SongFileDumper) -> Dumper:
    def dump(song: jbt.Song, path: Path, **kwargs: dict) -> Dict[Path, bytes]:
        name_format = FileNameFormat(Path("{title}.memon"), suggestion=path)
        songfile = internal_dumper(song, **kwargs)
        filepath = name_format.available_filename_for(songfile)
        return {filepath: songfile.contents}

    return dump



from typing import Any, Callable, Union

import click


def loader_option(*args: Any, **kwargs: Any) -> Callable:
    return click.option(
        *args, callback=add_to_dict("loader_options"), expose_value=False, **kwargs
    )


def dumper_option(*args: Any, **kwargs: Any) -> Callable:
    return click.option(
        *args, callback=add_to_dict("dumper_options"), expose_value=False, **kwargs
    )


def add_to_dict(
    key: str,
) -> Callable[[click.Context, Union[click.Option, click.Parameter], Any], None]:
    def add_to_key(
        ctx: click.Context, param: Union[click.Option, click.Parameter], value: Any
    ) -> None:
        # Avoid shadowing load/dump functions kwargs default values with the
        # default values chosen by click
        assert param.name is not None
        if not parameter_is_a_click_default(ctx, param.name):
            ctx.params.setdefault(key, {})[param.name] = value

    return add_to_key


def parameter_is_a_click_default(
    ctx: click.Context,
    name: str,
) -> bool:
    return ctx.get_parameter_source(name) in (
        click.core.ParameterSource.DEFAULT,
        click.core.ParameterSource.DEFAULT_MAP,
    )



from decimal import Decimal
from fractions import Fraction
from pathlib import Path
from typing import Dict, Iterable, Optional, Protocol, TypeVar, Union

from jubeatools import song

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)


class FileLoader(Protocol[T_co]):
    """Function that excepts a path to a file as a parameter and returns its
    contents in whatever form suitable for the current format. Returns None in
    case of error"""

    def __call__(self, path: Path) -> Optional[T_co]:
        ...


class FolderLoader(Protocol[T]):
    """Function that expects a folder or a file path as a parameter. Loads
    either all valid files in the folder or just the given file depending on
    the argument"""

    def __call__(self, path: Path) -> Dict[Path, T]:
        ...


def make_folder_loader(
    glob_pattern: str, file_loader: FileLoader[T]
) -> FolderLoader[T]:
    def folder_loader(path: Path) -> Dict[Path, T]:
        files: Dict[Path, T] = {}
        if path.is_dir():
            paths: Iterable[Path] = path.glob(glob_pattern)
        else:
            paths = [path]

        for p in paths:
            value = file_loader(p)
            if value is not None:
                files[p] = value

        return files

    return folder_loader


# TODO
# use numbers.Number instead when this mypy issue is finally fixed
# https://github.com/python/mypy/issues/3186
Number = Union[int, float, Decimal, Fraction]


def round_beats(beats: Number, denominator: int = 240) -> song.BeatsTime:
    """Rounds a given beat value to the nearest given division (default to
    nearest 1/240th"""
    nearest = round(beats * denominator)
    return song.BeatsTime(nearest, denominator)



"""Collection of parsing tools that are common to all the jubeat analyser formats"""

import re
import warnings
from collections import Counter
from copy import deepcopy
from dataclasses import astuple, dataclass
from decimal import Decimal
from itertools import product, zip_longest
from pathlib import Path
from typing import AbstractSet, Dict, List, Optional, Set, Tuple, Union

import constraint
from parsimonious import Grammar, NodeVisitor, ParseError
from parsimonious.nodes import Node

from jubeatools.formats.load_tools import make_folder_loader, round_beats
from jubeatools.song import BeatsTime, BPMEvent, Difficulty, LongNote, NotePosition

from .symbols import (
    CIRCLE_FREE_SYMBOLS,
    LONG_ARROW_DOWN,
    LONG_ARROW_LEFT,
    LONG_ARROW_RIGHT,
    LONG_ARROW_UP,
    NOTE_SYMBOLS,
)

DIFFICULTIES = {
    1: Difficulty.BASIC.value,
    2: Difficulty.ADVANCED.value,
    3: Difficulty.EXTREME.value,
}

SYMBOL_TO_BEATS_TIME = {c: BeatsTime("1/4") * i for i, c in enumerate(NOTE_SYMBOLS)}

CIRCLE_FREE_TO_BEATS_TIME = {
    c: BeatsTime("1/4") * i for i, c in enumerate(CIRCLE_FREE_SYMBOLS)
}

CIRCLE_FREE_TO_NOTE_SYMBOL = dict(zip(CIRCLE_FREE_SYMBOLS, NOTE_SYMBOLS))

LONG_ARROWS = LONG_ARROW_LEFT | LONG_ARROW_DOWN | LONG_ARROW_UP | LONG_ARROW_RIGHT

LONG_DIRECTION = {
    **{c: (1, 0) for c in LONG_ARROW_RIGHT},
    **{c: (-1, 0) for c in LONG_ARROW_LEFT},
    **{c: (0, 1) for c in LONG_ARROW_DOWN},
    **{c: (0, -1) for c in LONG_ARROW_UP},
}


EMPTY_LINE = re.compile(r"\s*(//.*)?")


# Any unicode character that's both :
#  - confusable with a dash/hyphen
#  - encodable in shift-jis-2004
# Gets added to the list of characters to be ignored in the timing section
EMPTY_BEAT_SYMBOLS = {
    "一",  # U+4E00 - CJK UNIFIED IDEOGRAPH-4E00
    "－",  # U+FF0D - FULLWIDTH HYPHEN-MINUS
    "ー",  # U+30FC - KATAKANA-HIRAGANA PROLONGED SOUND MARK
    "─",  # U+2500 - BOX DRAWINGS LIGHT HORIZONTAL
    "―",  # U+2015 - HORIZONTAL BAR
    "━",  # U+2501 - BOX DRAWINGS HEAVY HORIZONTAL
    "–",  # U+2013 - EN DASH
    "‐",  # U+2010 - HYPHEN
    "-",  # U+002D - HYPHEN-MINUS
    "−",  # U+2212 - MINUS SIGN
}


SEPARATOR = re.compile(r"--.*")


def is_separator(line: str) -> bool:
    return bool(SEPARATOR.match(line))


double_column_chart_line_grammar = Grammar(
    r"""
    line            = ws position_part ws (timing_part ws)? comment?
    position_part   = ~r"[^*#:|/\s]{4,8}"
    timing_part     = "|" ~r"[^*#:|/\s]*" "|"
    ws              = ~r"[\t \u3000]*"
    comment         = ~r"//.*"
"""
)


@dataclass
class DoubleColumnChartLine:
    position: str
    timing: Optional[str]

    def __str__(self) -> str:
        return f"{self.position} |{self.timing}|"

    def raise_if_unfit(self, bytes_per_panel: int) -> None:
        self.raise_if_position_unfit(bytes_per_panel)
        self.raise_if_timing_unfit(bytes_per_panel)

    def raise_if_position_unfit(self, bytes_per_panel: int) -> None:
        expected_length = 4 * bytes_per_panel
        actual_length = len(
            self.position.encode("shift-jis-2004", errors="surrogateescape")
        )
        if expected_length != actual_length:
            raise SyntaxError(
                f"Invalid position part. Since #bpp={bytes_per_panel}, the "
                f"position part of a line should be {expected_length} bytes long, "
                f"but {self.position!r} is {actual_length} bytes long"
            )

    def raise_if_timing_unfit(self, bytes_per_panel: int) -> None:
        if self.timing is None:
            return

        length = len(self.timing.encode("shift-jis-2004", errors="surrogateescape"))
        if length % bytes_per_panel != 0:
            raise SyntaxError(
                f"Invalid timing part. Since #bpp={bytes_per_panel}, the timing "
                f"part of a line should be divisible by {bytes_per_panel}, but "
                f"{self.timing!r} is {length} bytes long so it's not"
            )


class DoubleColumnChartLineVisitor(NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.pos_part: Optional[str] = None
        self.time_part: Optional[str] = None

    def visit_line(
        self, node: Node, visited_children: List[Node]
    ) -> DoubleColumnChartLine:
        if self.pos_part is None:
            raise ValueError("No positional part found after parsing line")
        return DoubleColumnChartLine(self.pos_part, self.time_part)

    def visit_position_part(self, node: Node, visited_children: List[Node]) -> None:
        self.pos_part = node.text

    def visit_timing_part(self, node: Node, visited_children: List[Node]) -> None:
        _, time_part, _ = node.children
        self.time_part = time_part.text

    def generic_visit(self, node: Node, visited_children: List[Node]) -> None:
        ...


def is_double_column_chart_line(line: str) -> bool:
    try:
        double_column_chart_line_grammar.parse(line)
    except ParseError:
        return False
    else:
        return True


def parse_double_column_chart_line(line: str) -> DoubleColumnChartLine:
    return DoubleColumnChartLineVisitor().visit(double_column_chart_line_grammar.parse(line))  # type: ignore


def is_empty_line(line: str) -> bool:
    return bool(EMPTY_LINE.fullmatch(line))


def split_double_byte_line(line: str) -> List[str]:
    """Split a #bpp=2 chart line into symbols.
    For example, Assuming "25" was defined as a symbol earlier :
    >>> split_chart_line("25口口25")
    ... ["25","口","口","25"]
    >>> split_chart_line("口⑪①25")
    ... ["口","⑪","①","25"]
    """
    encoded_line = line.encode("shift-jis-2004", errors="surrogateescape")
    if len(encoded_line) % 2 != 0:
        raise ValueError(
            "Line of odd length encountered while trying to split a double-byte "
            f"line : {line!r}"
        )
    symbols = []
    for i in range(0, len(encoded_line), 2):
        symbols.append(
            encoded_line[i : i + 2].decode("shift-jis-2004", errors="surrogateescape")
        )
    return symbols


@dataclass(frozen=True)
class UnfinishedLongNote:
    time: BeatsTime
    position: NotePosition
    tail_tip: NotePosition

    def ends_at(self, end: BeatsTime) -> LongNote:
        if end < self.time:
            raise ValueError(
                "Invalid end time. A long note starting at "
                f"{self.time} cannot end at {end} (which is earlier)"
            )
        return LongNote(
            time=self.time,
            position=self.position,
            duration=end - self.time,
            tail_tip=self.tail_tip,
        )


def find_long_note_candidates(
    bloc: List[List[str]],
    note_symbols: AbstractSet[str],
    should_skip: AbstractSet[NotePosition],
) -> Dict[NotePosition, Set[NotePosition]]:
    "Return a dict of arrow position to landing note candidates"
    arrow_to_note_candidates: Dict[NotePosition, Set[NotePosition]] = {}
    for y, x in product(range(4), range(4)):
        pos = NotePosition(x, y)
        if pos in should_skip:
            continue
        symbol = bloc[y][x]
        if symbol not in LONG_ARROWS:
            continue

        # at this point we are sure we have a long arrow
        # we need to check in its direction for note candidates
        note_candidates = set()
        𝛿pos = LONG_DIRECTION[symbol]
        candidate = NotePosition(x, y) + 𝛿pos
        while True:
            try:
                candidate = NotePosition.from_raw_position(candidate)
            except ValueError:
                break

            if candidate not in should_skip:
                new_symbol = bloc[candidate.y][candidate.x]
                if new_symbol in note_symbols:
                    note_candidates.add(candidate)
            candidate += 𝛿pos

        # if no notes have been crossed, we just ignore the arrow
        if note_candidates:
            arrow_to_note_candidates[pos] = note_candidates

    return arrow_to_note_candidates


Solution = Dict[NotePosition, NotePosition]
Candidates = Dict[NotePosition, Set[NotePosition]]


def pick_correct_long_note_candidates(
    arrow_to_note_candidates: Candidates,
    bloc: List[List[str]],
) -> Solution:
    """Believe it or not, assigning each arrow to a valid note candidate
    involves whipping out a CSP solver.
    Returns an arrow_pos -> note_pos mapping
    """
    problem = constraint.Problem()
    for arrow_pos, note_candidates in arrow_to_note_candidates.items():
        problem.addVariable(arrow_pos, list(note_candidates))
    problem.addConstraint(constraint.AllDifferentConstraint())
    solutions: List[Solution] = problem.getSolutions()
    if not solutions:
        raise SyntaxError(
            "Impossible arrow pattern found in block :\n"
            + "\n".join("".join(line) for line in bloc)
        )
    solution = min(solutions, key=long_note_solution_heuristic)
    if len(solutions) > 1 and not is_simple_solution(
        solution, arrow_to_note_candidates
    ):
        warnings.warn(
            "Ambiguous arrow pattern in bloc :\n"
            + "\n".join("".join(line) for line in bloc)
            + "\n"
            "The resulting long notes might not be what you expect"
        )
    return solution


def note_distance(a: NotePosition, b: NotePosition) -> float:
    return abs(complex(*astuple(a)) - complex(*astuple(b)))


def long_note_solution_heuristic(solution: Solution) -> Tuple[int, int, int]:
    c = Counter(int(note_distance(k, v)) for k, v in solution.items())
    return (c[3], c[2], c[1])


def is_simple_solution(solution: Solution, domains: Candidates) -> bool:
    return all(
        solution[v] == min(domains[v], key=lambda e: note_distance(e, v))
        for v in solution.keys()
    )


class JubeatAnalyserParser:
    def __init__(self) -> None:
        self.music: Optional[str] = None
        self.symbols = deepcopy(SYMBOL_TO_BEATS_TIME)
        self.section_starting_beat = BeatsTime(0)
        self.current_tempo = Decimal(120)
        self.timing_events: List[BPMEvent] = []
        self.offset = 0
        self.beats_per_section = BeatsTime(4)
        self.bytes_per_panel = 2
        self.level = Decimal(1)
        self.difficulty: Optional[str] = None
        self.title: Optional[str] = None
        self.artist: Optional[str] = None
        self.jacket: Optional[str] = None
        self.preview_start: Optional[int] = None
        self.hold_by_arrow = False
        self.circle_free = False

    def handle_command(self, command: str, value: Optional[str] = None) -> None:
        try:
            method = getattr(self, f"do_{command}")
        except AttributeError:
            raise SyntaxError(f"Unknown jubeat analyser command : {command}") from None

        if value is not None:
            method(value)
        else:
            method()

    def do_b(self, value: str) -> None:
        self.beats_per_section = round_beats(Decimal(value))

    def do_m(self, value: str) -> None:
        self.music = value

    def do_o(self, value: str) -> None:
        self.offset = int(value)

    def do_r(self, value: str) -> None:
        self.offset += int(value)

    def do_t(self, value: str) -> None:
        self.current_tempo = Decimal(value)
        self.timing_events.append(
            BPMEvent(
                time=self._current_beat(),
                BPM=self.current_tempo,
            )
        )

    def do_pw(self, value: str) -> None:
        if int(value) != 4:
            raise ValueError("jubeatools only supports 4×4 charts")

    do_ph = do_pw

    def do_lev(self, value: str) -> None:
        self.level = Decimal(value)

    def do_dif(self, value: str) -> None:
        dif = int(value)
        if dif < 4:
            self.difficulty = DIFFICULTIES[dif]
        else:
            self.difficulty = f"EDIT"

    def do_title(self, value: str) -> None:
        self.title = value

    def do_artist(self, value: str) -> None:
        self.artist = value

    def do_jacket(self, value: str) -> None:
        self.jacket = value

    def do_prevpos(self, value: str) -> None:
        self.preview_start = int(value)

    def _do_bpp(self, value: Union[int, str]) -> None:
        bpp = int(value)
        if bpp not in (1, 2):
            raise ValueError(f"Unexcpected bpp value : {value}")
        elif self.circle_free and bpp == 1:
            raise ValueError("Can't set #bpp to 1 when #circlefree is on")
        else:
            self.bytes_per_panel = bpp

    def do_holdbyarrow(self, value: str) -> None:
        self.hold_by_arrow = int(value) == 1

    def do_holdbytilde(self, value: str) -> None:
        raise NotImplementedError("jubeatools does not support #holdbytilde")

    def do_circlefree(self, raw_value: str) -> None:
        activate = bool(int(raw_value))
        if activate and self.bytes_per_panel != 2:
            raise ValueError("#circlefree can only be on when #bpp=2")
        self.circle_free = activate

    def _wrong_format(self, f: str) -> None:
        raise ValueError(
            f"{f} command means that this file uses another jubeat analyser "
            "format than the one the currently selected parser is designed for"
        )

    def do_memo(self) -> None:
        self._wrong_format("#memo")

    def do_memo1(self) -> None:
        self._wrong_format("#memo1")

    def do_boogie(self) -> None:
        self._wrong_format("#boogie")

    def do_memo2(self) -> None:
        self._wrong_format("#memo2")

    def define_symbol(self, symbol: str, timing: Decimal) -> None:
        bpp = self.bytes_per_panel
        length_as_shift_jis = len(
            symbol.encode("shift-jis-2004", errors="surrogateescape")
        )
        if length_as_shift_jis != bpp:
            raise ValueError(
                f"Invalid symbol definition. Since #bpp={bpp}, timing symbols "
                f"should be {bpp} bytes long but '{symbol}' is {length_as_shift_jis}"
            )
        if timing > self.beats_per_section:
            raise ValueError(
                f"Invalid symbol definition. Since sections only last "
                f"{self.beats_per_section} beats, a symbol cannot happen "
                f"afterwards at {timing}"
            )
        self.symbols[symbol] = round_beats(timing)

    def is_short_line(self, line: str) -> bool:
        return (
            len(line.encode("shift-jis-2004", errors="surrogateescape"))
            < self.bytes_per_panel * 4
        )

    def _split_chart_line(self, line: str) -> List[str]:
        if self.bytes_per_panel == 2:
            return split_double_byte_line(line)
        else:
            return list(line)

    def raise_if_separator(self, line: str, format_: str) -> None:
        if is_separator(line):
            raise SyntaxError(
                'Found a separator line (starting with "--") but the file '
                f"indicates it's using {format_} format, if the file is actually "
                f"in mono-column format (1列形式) there should be no {format_} line"
            )

    def _current_beat(self) -> BeatsTime:
        raise NotImplementedError


@dataclass
class DoubleColumnFrame:
    position_part: List[List[str]]
    timing_part: List[List[str]]

    def __str__(self) -> str:
        res = []
        for pos, time in zip_longest(self.position_part, self.timing_part):
            line = [f"{''.join(pos)}"]
            if time is not None:
                line += [f"|{''.join(time)}|"]
            res += [" ".join(line)]
        return "\n".join(res)


def read_jubeat_analyser_file(path: Path) -> Optional[List[str]]:
    """The vast majority of memo files you will encounter will be propely
    decoded using shift-jis-2004. Some won't but jubeat analyser works at the
    byte level so it doesn't care, here we use surrogateescape to handle
    potential decoding errors"""
    return path.read_text(encoding="shift-jis-2004", errors="surrogateescape").split(
        "\n"
    )


load_folder = make_folder_loader(
    glob_pattern="*.txt",
    file_loader=read_jubeat_analyser_file,
)



from dataclasses import dataclass

from jubeatools.song import Chart, Song


@dataclass
class JubeatFile:
    contents: bytes


@dataclass
class SongFile(JubeatFile):
    """File representing a collection of charts with metadata,
    like a .memon file for example"""

    song: Song


@dataclass
class ChartFile(SongFile):
    """File representing a single chart (with possibly some metadata),
    Used in jubeat analyser formats for instance"""

    difficulty: str
    chart: Chart



"""
memo

memo is the first (and probably oldest) youbeat-like format jubeat analyser
supports, the japanese docs give a good overview of what it looks like :

http://yosh52.web.fc2.com/jubeat/fumenformat.html

A chart in this format needs to have a `#memo` line somewhere to indicate its format

Like youbeat chart files, the position and timing information is split between
the left and right columns. However the timing column works differently.
In this precise format, the symbols in the timing part must only be interpeted
as quarter notes, even in the rare cases where a timing line is shorter than 4
symbols. It can never be longer than 4 either.
"""

from .dump import dump_memo
from .load import load_memo



from jubeatools.testutils.resources import load_test_file

from . import data


def test_MilK() -> None:
    """An actual malody chart will have may keys unspecified in the marshmallow
    schema, these should be ignored"""
    load_test_file(data, "1533908574.mc")



from fractions import Fraction

from jubeatools.song import LongNote, NotePosition, TapNote

notes = {
    TapNote(time=Fraction(0, 1), position=NotePosition(x=0, y=0)),
    TapNote(time=Fraction(0, 1), position=NotePosition(x=0, y=1)),
    TapNote(time=Fraction(0, 1), position=NotePosition(x=0, y=3)),
    LongNote(
        time=Fraction(0, 1),
        position=NotePosition(x=1, y=0),
        duration=Fraction(1, 16),
        tail_tip=NotePosition(x=2, y=0),
    ),
    TapNote(time=Fraction(0, 1), position=NotePosition(x=1, y=1)),
    LongNote(
        time=Fraction(0, 1),
        position=NotePosition(x=1, y=2),
        duration=Fraction(1, 3),
        tail_tip=NotePosition(x=2, y=2),
    ),
    LongNote(
        time=Fraction(1, 8),
        position=NotePosition(x=0, y=3),
        duration=Fraction(1, 4),
        tail_tip=NotePosition(x=2, y=3),
    ),
    TapNote(time=Fraction(1, 4), position=NotePosition(x=0, y=0)),
    TapNote(time=Fraction(5, 16), position=NotePosition(x=1, y=0)),
    LongNote(
        time=Fraction(1, 2),
        position=NotePosition(x=0, y=0),
        duration=Fraction(1, 16),
        tail_tip=NotePosition(x=1, y=0),
    ),
    LongNote(
        time=Fraction(5, 8),
        position=NotePosition(x=0, y=3),
        duration=Fraction(1, 4),
        tail_tip=NotePosition(x=1, y=3),
    ),
    TapNote(time=Fraction(13, 16), position=NotePosition(x=0, y=0)),
    LongNote(
        time=Fraction(17, 16),
        position=NotePosition(x=0, y=0),
        duration=Fraction(1, 4),
        tail_tip=NotePosition(x=1, y=0),
    ),
    LongNote(
        time=Fraction(25, 16),
        position=NotePosition(x=0, y=0),
        duration=Fraction(1, 4),
        tail_tip=NotePosition(x=2, y=0),
    ),
    TapNote(time=Fraction(33, 16), position=NotePosition(x=0, y=0)),
    TapNote(time=Fraction(57, 8), position=NotePosition(x=1, y=1)),
    TapNote(time=Fraction(59, 8), position=NotePosition(x=1, y=1)),
    TapNote(time=Fraction(61, 8), position=NotePosition(x=1, y=1)),
    TapNote(time=Fraction(63, 8), position=NotePosition(x=1, y=1)),
    TapNote(time=Fraction(79, 8), position=NotePosition(x=1, y=1)),
    TapNote(time=Fraction(187, 16), position=NotePosition(x=1, y=0)),
    TapNote(time=Fraction(191, 16), position=NotePosition(x=1, y=0)),
    LongNote(
        time=Fraction(97, 8),
        position=NotePosition(x=1, y=1),
        duration=Fraction(1, 4),
        tail_tip=NotePosition(x=2, y=1),
    ),
    TapNote(time=Fraction(195, 16), position=NotePosition(x=1, y=0)),
}



from typing import Dict

from . import jubeat_analyser, konami, malody, memon
from .format_names import Format
from .typing import Dumper, Loader

#: Maps each Format enum member to its associated loader
LOADERS: Dict[Format, Loader] = {
    Format.EVE: konami.load_eve,
    Format.JBSQ: konami.load_jbsq,
    Format.MALODY: malody.load_malody,
    Format.MEMON_LEGACY: memon.load_memon_legacy,
    Format.MEMON_0_1_0: memon.load_memon_0_1_0,
    Format.MEMON_0_2_0: memon.load_memon_0_2_0,
    Format.MEMON_0_3_0: memon.load_memon_0_3_0,
    Format.MEMON_1_0_0: memon.load_memon_1_0_0,
    Format.MONO_COLUMN: jubeat_analyser.load_mono_column,
    Format.MEMO: jubeat_analyser.load_memo,
    Format.MEMO_1: jubeat_analyser.load_memo1,
    Format.MEMO_2: jubeat_analyser.load_memo2,
}

#: Maps each Format enum member to its associated dumper
DUMPERS: Dict[Format, Dumper] = {
    Format.EVE: konami.dump_eve,
    Format.JBSQ: konami.dump_jbsq,
    Format.MALODY: malody.dump_malody,
    Format.MEMON_LEGACY: memon.dump_memon_legacy,
    Format.MEMON_0_1_0: memon.dump_memon_0_1_0,
    Format.MEMON_0_2_0: memon.dump_memon_0_2_0,
    Format.MEMON_0_3_0: memon.dump_memon_0_3_0,
    Format.MEMON_1_0_0: memon.dump_memon_1_0_0,
    Format.MONO_COLUMN: jubeat_analyser.dump_mono_column,
    Format.MEMO: jubeat_analyser.dump_memo,
    Format.MEMO_1: jubeat_analyser.dump_memo1,
    Format.MEMO_2: jubeat_analyser.dump_memo2,
}



""".eve is the file format used in arcade releases of jubeat

.eve files are CSVs with three columns : time, command, value

A small but annoying amount of precision is lost when using this format :
- time is stored already "rendered" as a whole number of ticks on a 300Hz clock
  instead of using symbolic time
- while some symbolic time information remains in the form of TEMPO, MEASURE
  and HAKU commands (respectively a BPM change, a measure marker and a beat
  marker), BPMs are stored as `int((6*10^7)/BPM)` which makes it hard to
  recover many significant digits
"""

from .dump import dump_eve
from .load import load_eve



"""The JBSQ format described using construct.
see https://construct.readthedocs.io/en/latest/index.html"""

from dataclasses import dataclass
from typing import List, Optional

import construct as c
import construct_typed as ct


class EventType(ct.EnumBase):
    PLAY = 1
    END = 2
    MEASURE = 3
    HAKU = 4
    TEMPO = 5
    LONG = 6


@dataclass
class Event(ct.TContainerMixin):
    type_: EventType = ct.sfield(ct.TEnum(c.Byte, EventType))
    time_in_ticks: int = ct.sfield(c.Int24ul)
    value: int = ct.sfield(c.Int32ul)


@dataclass
class JBSQ(ct.TContainerMixin):
    magic: Optional[bytes] = ct.sfield(
        c.Select(c.Const(b"IJBQ"), c.Const(b"IJSQ"), c.Const(b"JBSQ"))
    )
    num_events: int = ct.sfield(c.Int32ul)
    combo: int = ct.sfield(c.Int32ul)
    end_time: int = ct.sfield(c.Int32ul)
    _1: None = ct.sfield(c.Padding(2))
    starting_buttons: int = ct.sfield(c.Int16ul)
    start_time: int = ct.sfield(c.Int32ul)
    _2: None = ct.sfield(c.Padding(12))
    density_graph: List[int] = ct.sfield(c.Byte[60])
    events: List[Event] = ct.sfield(c.Array(c.this.num_events, ct.TStruct(Event)))


jbsq = ct.TStruct(JBSQ)



"""
memo2

memo2 is the jubeat analyser version of youbeat files :

http://yosh52.web.fc2.com/jubeat/fumenformat.html

A chart in this format needs to have a `#memo2` line somewhere to indicate its format
"""

from .dump import dump_memo2
from .load import load_memo2



"""General utility functions"""

import unicodedata
from collections import defaultdict
from decimal import Decimal
from fractions import Fraction
from functools import reduce
from math import gcd
from typing import Callable, Dict, Hashable, Iterable, List, Optional, TypeVar


def single_lcm(a: int, b: int) -> int:
    """Return lowest common multiple of two numbers"""
    return a * b // gcd(a, b)


def lcm(*args: int) -> int:
    """Return lcm of args."""
    return reduce(single_lcm, args, 1)


def charinfo(c: str) -> str:
    """Return some info on the character"""
    return f"{c!r}  # U+{ord(c):05X} : {unicodedata.name(c)}"


A = TypeVar("A")
B = TypeVar("B")

# Monadic stuff !
def none_or(c: Callable[[A], B], e: Optional[A]) -> Optional[B]:
    if e is None:
        return None
    else:
        return c(e)


def fraction_to_decimal(frac: Fraction) -> Decimal:
    "Thanks stackoverflow ! https://stackoverflow.com/a/40468867/10768117"
    return frac.numerator / Decimal(frac.denominator)


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


def group_by(elements: Iterable[V], key: Callable[[V], K]) -> Dict[K, List[V]]:
    res = defaultdict(list)
    for e in elements:
        res[key(e)].append(e)

    return res



import math

from hypothesis import given
from hypothesis import strategies as st

from jubeatools.formats.konami.commons import bpm_to_value, value_to_truncated_bpm


@given(st.integers(min_value=1, max_value=6 * 10**7))
def test_that_truncating_preserves_tempo_value(original_value: int) -> None:
    truncated_bpm = value_to_truncated_bpm(original_value)
    raw_recovered_value = bpm_to_value(truncated_bpm)
    recovered_value = math.floor(raw_recovered_value)
    assert recovered_value == original_value



""".jbsq is the file format used in the iOS and Android app Jubeat plus

It's a binary version of the .eve format, with the same limitations.

I wanted to try kaitai for this but it doesn't support serializing right now"""

from .dump import dump_jbsq
from .load import load_jbsq



from copy import deepcopy
from dataclasses import astuple, dataclass
from decimal import Decimal
from itertools import product
from pathlib import Path
from typing import Any, Dict, Iterator, List, Set, Tuple, Union

from parsimonious import Grammar, NodeVisitor, ParseError
from parsimonious.nodes import Node

from jubeatools.song import (
    BeatsTime,
    Chart,
    LongNote,
    Metadata,
    NotePosition,
    Preview,
    SecondsTime,
    Song,
    TapNote,
    Timing,
)
from jubeatools.utils import none_or

from ..command import is_command, parse_command
from ..load_tools import (
    CIRCLE_FREE_TO_BEATS_TIME,
    JubeatAnalyserParser,
    UnfinishedLongNote,
    find_long_note_candidates,
    is_empty_line,
    is_separator,
    load_folder,
    pick_correct_long_note_candidates,
    split_double_byte_line,
)
from ..symbol_definition import is_symbol_definition, parse_symbol_definition
from ..symbols import CIRCLE_FREE_SYMBOLS

mono_column_chart_line_grammar = Grammar(
    r"""
    line            = ws chart_line ws comment?
    chart_line      = ~r"[^*#:|\-/\s]{4,8}"
    ws              = ~r"[\t \u3000]*"
    comment         = ~r"//.*"
"""
)


class MonoColumnChartLineVisitor(NodeVisitor):
    def visit_line(self, node: Node, visited_children: List[Node]) -> str:
        _, chart_line, _, _ = node.children
        return chart_line.text  # type: ignore

    def generic_visit(self, node: Node, visited_children: list) -> None:
        ...


def is_mono_column_chart_line(line: str) -> bool:
    try:
        mono_column_chart_line_grammar.parse(line)
    except ParseError:
        return False
    else:
        return True


def parse_mono_column_chart_line(line: str) -> str:
    return MonoColumnChartLineVisitor().visit(  # type: ignore
        mono_column_chart_line_grammar.parse(line)
    )


@dataclass
class MonoColumnLoadedSection:
    """
    An intermediate reprensetation of what's parsed from a mono-column
    chart section, contains :
    - raw chart lines
    - defined timing symbols
    - length in beats (usually 4)
    - tempo
    """

    chart_lines: List[str]
    symbols: Dict[str, BeatsTime]
    length: BeatsTime
    tempo: Decimal

    def blocs(self, bpp: int = 2) -> Iterator[List[List[str]]]:
        if bpp not in (1, 2):
            raise ValueError(f"Invalid bpp : {bpp}")
        elif bpp == 2:
            split_line = split_double_byte_line
        else:
            split_line = lambda l: list(l)

        for i in range(0, len(self.chart_lines), 4):
            yield [split_line(self.chart_lines[i + j]) for j in range(4)]


class MonoColumnParser(JubeatAnalyserParser):
    def __init__(self) -> None:
        super().__init__()
        self.current_chart_lines: List[str] = []
        self.sections: List[MonoColumnLoadedSection] = []

    def _current_beat(self) -> BeatsTime:
        return self.section_starting_beat

    def do_bpp(self, value: Union[int, str]) -> None:
        if self.sections:
            raise ValueError(
                "This file apparently changes its bytes per panel value (#bpp) "
                "halfway through the chart, jubeatools does not handle that"
            )
        else:
            self._do_bpp(value)

    def move_to_next_section(self) -> None:
        if len(self.current_chart_lines) % 4 != 0:
            raise SyntaxError("Current section is missing chart lines")
        else:
            self.sections.append(
                MonoColumnLoadedSection(
                    chart_lines=self.current_chart_lines,
                    symbols=deepcopy(self.symbols),
                    length=self.beats_per_section,
                    tempo=self.current_tempo,
                )
            )
            self.current_chart_lines = []
            self.section_starting_beat += self.beats_per_section

    def append_chart_line(self, line: str) -> None:
        expected_length = 4 * self.bytes_per_panel
        actual_length = len(line.encode("shift-jis-2004", errors="surrogateescape"))
        if actual_length != expected_length:
            raise SyntaxError(f"Invalid chart line. Since for ")
        if self.bytes_per_panel == 1 and len(line) != 4:
            raise SyntaxError(f"Invalid chart line for #bpp=1 : {line}")
        elif (
            self.bytes_per_panel == 2
            and len(line.encode("shift-jis-2004", errors="surrogateescape")) != 8
        ):
            raise SyntaxError(f"Invalid chart line for #bpp=2 : {line}")
        self.current_chart_lines.append(line)

    def load_line(self, raw_line: str) -> None:
        line = raw_line.strip()
        if is_command(line):
            command, value = parse_command(line)
            self.handle_command(command, value)
        elif is_symbol_definition(line):
            symbol, timing = parse_symbol_definition(line)
            self.define_symbol(symbol, timing)
        elif is_mono_column_chart_line(line):
            chart_line = parse_mono_column_chart_line(line)
            self.append_chart_line(chart_line)
        elif is_separator(line):
            self.move_to_next_section()
        elif not (is_empty_line(line) or self.is_short_line(line)):
            raise SyntaxError(f"not a valid mono-column file line : {line}")

    def notes(self) -> Iterator[Union[TapNote, LongNote]]:
        if self.hold_by_arrow:
            yield from self._iter_notes()
        else:
            yield from self._iter_notes_without_longs()

    def _iter_blocs(
        self,
    ) -> Iterator[Tuple[BeatsTime, MonoColumnLoadedSection, List[List[str]]]]:
        section_starting_beat = BeatsTime(0)
        for section in self.sections:
            for bloc in section.blocs():
                yield section_starting_beat, section, bloc
            section_starting_beat += section.length

    def _iter_notes(self) -> Iterator[Union[TapNote, LongNote]]:
        unfinished_longs: Dict[NotePosition, UnfinishedLongNote] = {}
        for section_starting_beat, section, bloc in self._iter_blocs():
            should_skip: Set[NotePosition] = set()
            # 1/3 : look for ends to unfinished long notes
            for pos, unfinished_long in unfinished_longs.items():
                x, y = astuple(pos)
                symbol = bloc[y][x]
                if self.circle_free and symbol in CIRCLE_FREE_SYMBOLS:
                    should_skip.add(pos)
                    symbol_time = CIRCLE_FREE_TO_BEATS_TIME[symbol]
                    note_time = section_starting_beat + symbol_time
                    yield unfinished_long.ends_at(note_time)
                elif symbol in section.symbols:
                    should_skip.add(pos)
                    symbol_time = section.symbols[symbol]
                    note_time = section_starting_beat + symbol_time
                    yield unfinished_long.ends_at(note_time)

            unfinished_longs = {
                k: unfinished_longs[k] for k in unfinished_longs.keys() - should_skip
            }

            # 2/3 : look for new long notes starting on this bloc
            arrow_to_note_candidates = find_long_note_candidates(
                bloc, section.symbols.keys(), should_skip
            )
            if arrow_to_note_candidates:
                solution = pick_correct_long_note_candidates(
                    arrow_to_note_candidates,
                    bloc,
                )
                for arrow_pos, note_pos in solution.items():
                    should_skip.add(arrow_pos)
                    should_skip.add(note_pos)
                    symbol = bloc[note_pos.y][note_pos.x]
                    symbol_time = section.symbols[symbol]
                    note_time = section_starting_beat + symbol_time
                    unfinished_longs[note_pos] = UnfinishedLongNote(
                        time=note_time, position=note_pos, tail_tip=arrow_pos
                    )

            # 3/3 : find regular notes
            for y, x in product(range(4), range(4)):
                position = NotePosition(x, y)
                if position in should_skip:
                    continue
                symbol = bloc[y][x]
                if symbol in section.symbols:
                    symbol_time = section.symbols[symbol]
                    note_time = section_starting_beat + symbol_time
                    yield TapNote(note_time, position)

    def _iter_notes_without_longs(self) -> Iterator[TapNote]:
        section_starting_beat = BeatsTime(0)
        for section in self.sections:
            for bloc, y, x in product(section.blocs(), range(4), range(4)):
                symbol = bloc[y][x]
                if symbol in section.symbols:
                    symbol_time = section.symbols[symbol]
                    note_time = section_starting_beat + symbol_time
                    position = NotePosition(x, y)
                    yield TapNote(note_time, position)
            section_starting_beat += section.length


def load_mono_column(path: Path, **kwargs: Any) -> Song:
    files = load_folder(path)
    charts = [_load_mono_column_file(lines) for _, lines in files.items()]
    return Song.from_monochart_instances(*charts)


def _load_mono_column_file(lines: List[str]) -> Song:
    parser = MonoColumnParser()
    for i, raw_line in enumerate(lines, start=1):
        try:
            parser.load_line(raw_line)
        except Exception as e:
            raise SyntaxError(f"On line {i}\n{e}")

    metadata = Metadata(
        title=parser.title,
        artist=parser.artist,
        audio=none_or(Path, parser.music),
        cover=none_or(Path, parser.jacket),
    )
    if parser.preview_start is not None:
        metadata.preview = Preview(
            start=SecondsTime(parser.preview_start) / 1000, length=SecondsTime(10)
        )

    timing = Timing(
        events=parser.timing_events, beat_zero_offset=SecondsTime(parser.offset) / 1000
    )
    charts = {
        parser.difficulty
        or "EXT": Chart(
            level=Decimal(parser.level),
            timing=timing,
            notes=sorted(parser.notes(), key=lambda n: (n.time, n.position)),
        )
    }
    return Song(metadata=metadata, charts=charts)



from pathlib import Path
from typing import Any, Dict, List, Protocol

from jubeatools.formats.filetypes import ChartFile, SongFile
from jubeatools.song import Song


# Dumpers take a Song object and a Path hint and give back a dict that maps
#
class Dumper(Protocol):
    """A Dumper is a callable that takes in a Song object, a Path hint and
    potential options, then gives back a dict that maps file name suggestions
    to the binary content of the file"""

    def __call__(self, song: Song, path: Path, **kwargs: Any) -> Dict[Path, bytes]:
        ...


class ChartFileDumper(Protocol):
    """Generic signature of internal dumper for formats that use one file
    per chart"""

    def __call__(self, song: Song, **kwargs: Any) -> List[ChartFile]:
        ...


class SongFileDumper(Protocol):
    """Generic signature of internal dumper for formats that use a single file
    to hold all charts of a song"""

    def __call__(self, song: Song, **kwargs: Any) -> SongFile:
        ...


class Loader(Protocol):
    """A Loader deserializes a Path to a Song object and possibly takes in
    some options via the kwargs.
    The Path can be a file or a folder depending on the format"""

    def __call__(self, path: Path, **kwargs: Any) -> Song:
        ...



"""
Mono column (not an official name)

It's the format jubeat analyser expects when otherwise no format command has
been found (like #memo #memo1 #memo2 or #boogie)

Mono-column files are usually properly decoded using `shift-jis-2004`

Mono-column files are made up of several sections.
Each section is made up of a series of command lines and chart lines.
The section end is marked by a line starting with "--"

Command lines follow the usual jubeat analyser command pattern

Chart lines come in groups of 4 in each section.
A group of 4 chart lines makes up a frame, which contains the note symbols.

Note symbols encode both the position and the time. The position is determined
"visually" by the position the note symbol occupies in the frame.
The time within the frame is determined by the precise symbol used.

The default note symbols are the circled numbers from ① to ⑯, they represent
"""

from .dump import dump_mono_column
from .load import load_mono_column



from __future__ import annotations

from dataclasses import dataclass, replace
from fractions import Fraction
from typing import List, Union

from more_itertools import windowed
from sortedcontainers import SortedKeyList

from jubeatools import song
from jubeatools.formats.load_tools import round_beats
from jubeatools.utils import fraction_to_decimal, group_by


@dataclass
class BPMAtSecond:
    seconds: Fraction
    BPM: Fraction


@dataclass
class BPMAtBeat:
    beats: Fraction
    BPM: Fraction


@dataclass
class SecondsAtBeat:
    seconds: Fraction
    beats: Fraction


@dataclass
class BPMChange:
    beats: song.BeatsTime
    seconds: Fraction
    BPM: Fraction


@dataclass
class TimeMap:
    """Wraps a song.Timing to allow converting symbolic time (in beats)
    to clock time (in seconds) and back"""

    events_by_beats: SortedKeyList[BPMChange, song.BeatsTime]
    events_by_seconds: SortedKeyList[BPMChange, Fraction]

    @classmethod
    def from_timing(cls, timing: song.Timing) -> TimeMap:
        """Create a time map from a song.Timing object"""
        return cls.from_beats(
            events=[
                BPMAtBeat(beats=e.time, BPM=Fraction(e.BPM)) for e in timing.events
            ],
            offset=SecondsAtBeat(
                seconds=Fraction(timing.beat_zero_offset), beats=Fraction(0)
            ),
        )

    @classmethod
    def from_beats(cls, events: List[BPMAtBeat], offset: SecondsAtBeat) -> TimeMap:
        """Create a time map from a list of BPM changes with times given in
        beats, the offset parameter is more flexible than a "regular" beat zero
        offset as it accepts non-zero beats"""
        if not events:
            raise ValueError("No BPM defined")

        grouped_by_time = group_by(events, key=lambda e: e.beats)
        for time, events_at_time in grouped_by_time.items():
            if len(events_at_time) > 1:
                raise ValueError(f"Multiple BPMs defined at beat {time} : {events}")

        # First compute everything as if the first BPM change happened at
        # zero seconds, then shift according to the offset
        sorted_events = sorted(events, key=lambda e: e.beats)
        first_event = sorted_events[0]
        current_second = Fraction(0)
        bpm_changes = [
            BPMChange(first_event.beats, current_second, Fraction(first_event.BPM))
        ]
        for previous, current in windowed(sorted_events, 2):
            if previous is None or current is None:
                continue

            beats_since_last_event = current.beats - previous.beats
            seconds_since_last_event = (60 * beats_since_last_event) / Fraction(
                previous.BPM
            )
            current_second += seconds_since_last_event
            bpm_change = BPMChange(current.beats, current_second, Fraction(current.BPM))
            bpm_changes.append(bpm_change)

        not_shifted = cls(
            events_by_beats=SortedKeyList(bpm_changes, key=lambda b: b.beats),
            events_by_seconds=SortedKeyList(bpm_changes, key=lambda b: b.seconds),
        )
        unshifted_seconds_at_offset = not_shifted.fractional_seconds_at(offset.beats)
        shift = offset.seconds - unshifted_seconds_at_offset
        shifted_bpm_changes = [
            replace(b, seconds=b.seconds + shift) for b in bpm_changes
        ]
        return cls(
            events_by_beats=SortedKeyList(shifted_bpm_changes, key=lambda b: b.beats),
            events_by_seconds=SortedKeyList(
                shifted_bpm_changes, key=lambda b: b.seconds
            ),
        )

    @classmethod
    def from_seconds(cls, events: List[BPMAtSecond]) -> TimeMap:
        """Create a time map from a list of BPM changes with time positions
        given in seconds. The first BPM implicitely happens at beat zero"""
        if not events:
            raise ValueError("No BPM defined")

        grouped_by_time = group_by(events, key=lambda e: e.seconds)
        for time, events_at_time in grouped_by_time.items():
            if len(events_at_time) > 1:
                raise ValueError(f"Multiple BPMs defined at {time} seconds : {events}")

        # take the first BPM change then compute from there
        sorted_events = sorted(events, key=lambda e: e.seconds)
        first_event = sorted_events[0]
        current_beat = Fraction(0)
        bpm_changes = [BPMChange(current_beat, first_event.seconds, first_event.BPM)]
        for previous, current in windowed(sorted_events, 2):
            if previous is None or current is None:
                continue

            seconds_since_last_event = current.seconds - previous.seconds
            beats_since_last_event = (
                previous.BPM * seconds_since_last_event
            ) / Fraction(60)
            current_beat += beats_since_last_event
            bpm_change = BPMChange(current_beat, current.seconds, current.BPM)
            bpm_changes.append(bpm_change)

        return cls(
            events_by_beats=SortedKeyList(bpm_changes, key=lambda b: b.beats),
            events_by_seconds=SortedKeyList(bpm_changes, key=lambda b: b.seconds),
        )

    def seconds_at(self, beat: song.BeatsTime) -> song.SecondsTime:
        frac_seconds = self.fractional_seconds_at(beat)
        return fraction_to_decimal(frac_seconds)

    def fractional_seconds_at(self, beat: song.BeatsTime) -> Fraction:
        """Before the first bpm change, compute backwards from the first bpm,
        after the first bpm change, compute forwards from the previous bpm
        change"""
        index = self.events_by_beats.bisect_key_right(beat)
        first_or_previous_index = max(0, index - 1)
        bpm_change: BPMChange = self.events_by_beats[first_or_previous_index]
        beats_since_last_event = beat - bpm_change.beats
        seconds_since_last_event = (60 * beats_since_last_event) / bpm_change.BPM
        return bpm_change.seconds + seconds_since_last_event

    def beats_at(self, seconds: Union[song.SecondsTime, Fraction]) -> song.BeatsTime:
        frac_seconds = Fraction(seconds)
        index = self.events_by_seconds.bisect_key_right(frac_seconds)
        first_or_previous_index = max(0, index - 1)
        bpm_change: BPMChange = self.events_by_seconds[first_or_previous_index]
        seconds_since_last_event = frac_seconds - bpm_change.seconds
        beats_since_last_event = (bpm_change.BPM * seconds_since_last_event) / Fraction(
            60
        )
        return bpm_change.beats + beats_since_last_event

    def convert_to_timing_info(self, beat_snap: int = 240) -> song.Timing:
        return song.Timing(
            events=[
                song.BPMEvent(
                    time=round_beats(e.beats, beat_snap),
                    BPM=fraction_to_decimal(e.BPM),
                )
                for e in self.events_by_beats
            ],
            beat_zero_offset=self.seconds_at(song.BeatsTime(0)),
        )



from hypothesis import given
from hypothesis import strategies as st

from jubeatools import song
from jubeatools.formats.konami.commons import EveLong
from jubeatools.formats.timemap import TimeMap
from jubeatools.testutils import strategies as jbst


@given(
    jbst.long_note(),
    jbst.timing_info(
        with_bpm_changes=True,
        bpm_strat=st.decimals(min_value=1, max_value=1000, places=2),
        beat_zero_offset_strat=st.decimals(min_value=0, max_value=20, places=2),
    ),
)
def test_that_long_note_roundtrips(
    long_note: song.LongNote, timing: song.Timing
) -> None:
    time_map = TimeMap.from_timing(timing)
    original = EveLong.from_jubeatools(long_note, time_map)
    recovered = EveLong.from_value(original.value)
    assert recovered == original



from dataclasses import fields
from decimal import Decimal
from typing import Optional

import simplejson as json
from hypothesis import given
from hypothesis import strategies as st

from jubeatools import song
from jubeatools.formats import Format
from jubeatools.formats.malody import schema as malody
from jubeatools.formats.malody.dump import dump_malody_chart
from jubeatools.testutils import strategies as jbst
from jubeatools.testutils.test_patterns import dump_and_load_then_compare


@st.composite
def difficulty(draw: st.DrawFn) -> str:
    diff: str = draw(st.sampled_from(list(d.value for d in song.Difficulty)))
    return diff


@st.composite
def chart(draw: st.DrawFn) -> song.Chart:
    c: song.Chart = draw(jbst.chart(level_strat=st.just(Decimal(0))))
    return c


@st.composite
def metadata(draw: st.DrawFn) -> song.Metadata:
    metadata: song.Metadata = draw(jbst.metadata())
    metadata.preview = None
    metadata.preview_file = None
    return metadata


@st.composite
def malody_song(draw: st.DrawFn) -> song.Song:
    """Malody files only hold one chart and have limited metadata"""
    diff = draw(difficulty())
    chart_ = draw(chart())
    metadata_ = draw(metadata())
    return song.Song(metadata=metadata_, charts={diff: chart_})


@given(malody_song())
def test_that_full_chart_roundtrips(s: song.Song) -> None:
    dump_and_load_then_compare(
        Format.MALODY,
        s,
        bytes_decoder=lambda b: b.decode("utf-8"),
    )


@given(chart(), metadata(), st.one_of(st.none(), difficulty()))
def test_that_none_values_in_metadata_dont_appear_in_dumped_json(
    chart: song.Chart,
    metadata: song.Metadata,
    dif: Optional[str],
) -> None:
    assert chart.timing is not None
    malody_chart = dump_malody_chart(metadata, dif, chart, chart.timing)
    json_chart = malody.CHART_SCHEMA.dump(malody_chart)
    assert all(value is not None for value in json_chart["meta"].values())


@given(malody_song())
def test_that_fields_are_ordered(s: song.Song) -> None:
    dif, chart = next(iter(s.charts.items()))
    assert chart.timing is not None
    malody_chart = dump_malody_chart(s.metadata, dif, chart, chart.timing)
    json_chart = malody.CHART_SCHEMA.dump(malody_chart)
    text_chart = json.dumps(json_chart, indent=4, use_decimal=True)
    reparsed_chart = json.loads(
        text_chart,
    )
    # dict is ordered in 3.8 ... right ?
    order_in_file = list(reparsed_chart["meta"].keys())
    order_in_definition = list(
        f.name for f in fields(malody.Metadata) if f.name in reparsed_chart["meta"]
    )
    assert order_in_file == order_in_definition



from typing import Iterable, Union

import pytest

from jubeatools.formats.jubeat_analyser.mono_column.load import MonoColumnParser
from jubeatools.song import BeatsTime, LongNote, NotePosition, TapNote


def compare_chart_notes(
    chart: str, expected: Iterable[Union[TapNote, LongNote]]
) -> None:
    parser = MonoColumnParser()
    for line in chart.split("\n"):
        parser.load_line(line)
    actual = list(parser.notes())
    assert set(expected) == set(actual)


def test_simple_section() -> None:
    chart = """
        ①□□□
        □⑤□□
        □□⑨□
        □□□⑬
        -------
        """
    expected = [
        TapNote(time=BeatsTime(i), position=NotePosition(i, i)) for i in range(4)
    ]
    compare_chart_notes(chart, expected)


def test_compound_section() -> None:
    chart = """
        □①①□
        □⑩⑪□
        ④⑧⑨⑤
        ③⑥⑦③

        ⑯⑫⑬⑯
        □□□□
        □□□□
        ⑭□□⑭
        ------------- 2
        """
    expected = [
        TapNote(time=BeatsTime("1/4") * (t - 1), position=NotePosition(x, y))
        for t, x, y in [
            (1, 1, 0),
            (1, 2, 0),
            (3, 0, 3),
            (3, 3, 3),
            (4, 0, 2),
            (5, 3, 2),
            (6, 1, 3),
            (7, 2, 3),
            (8, 1, 2),
            (9, 2, 2),
            (10, 1, 1),
            (11, 2, 1),
            (12, 1, 0),
            (13, 2, 0),
            (14, 0, 3),
            (14, 3, 3),
            (16, 0, 0),
            (16, 3, 0),
        ]
    ]
    compare_chart_notes(chart, expected)


def test_symbol_definition() -> None:
    chart = """
        *Ａ:2 //⑨と同タイミング
        *Ｂ:2.125
        *Ｃ:2.25 //⑩と同じ
        *Ｄ:2.375
        *Ｅ:2.5 //⑪と同じ
        *Ｆ:2.625
        *Ｇ:2.75 //⑫と同じ
        *Ｈ:2.875
        *Ｉ:3 //⑬と同じ
        *Ｊ:3.125
        *Ｋ:3.25 //⑭と同じ
        *Ｌ:3.375

        ＡＢＣＤ
        Ｌ③□Ｅ
        Ｋ⑦□Ｆ
        ＪＩＨＧ
        --
        """
    expected = [
        TapNote(BeatsTime(t), NotePosition(x, y))
        for t, x, y in [
            ("4/8", 1, 1),
            ("12/8", 1, 2),
            ("16/8", 0, 0),
            ("17/8", 1, 0),
            ("18/8", 2, 0),
            ("19/8", 3, 0),
            ("20/8", 3, 1),
            ("21/8", 3, 2),
            ("22/8", 3, 3),
            ("23/8", 2, 3),
            ("24/8", 1, 3),
            ("25/8", 0, 3),
            ("26/8", 0, 2),
            ("27/8", 0, 1),
        ]
    ]
    compare_chart_notes(chart, expected)


def test_half_width_symbols() -> None:
    chart = """
        b=7
        *⑲:4.5
        *21:5
        *25:6

        口⑪①口
        ①⑮⑤⑪
        ⑤口口⑮
        ⑨口口⑨

        21口口⑲
        25口口25
        口⑲21口
        25口口25
        ----------
        """
    expected = [
        TapNote(BeatsTime(t), NotePosition(x, y))
        for t, x, y in [
            ("0.0", 2, 0),
            ("0.0", 0, 1),
            ("1.0", 2, 1),
            ("1.0", 0, 2),
            ("2.0", 0, 3),
            ("2.0", 3, 3),
            ("2.5", 1, 0),
            ("2.5", 3, 1),
            ("3.5", 1, 1),
            ("3.5", 3, 2),
            ("4.5", 1, 2),
            ("4.5", 3, 0),
            ("5.0", 0, 0),
            ("5.0", 2, 2),
            ("6.0", 0, 1),
            ("6.0", 3, 1),
            ("6.0", 0, 3),
            ("6.0", 3, 3),
        ]
    ]
    compare_chart_notes(chart, expected)


def test_irregular_beats_per_frame_1() -> None:
    chart = """
        b=2.75
        ①□□□
        □□□□
        □□□□
        □□□□
        --
        □□□□
        □①□□
        □□□□
        □□□□
        --
        □□□□
        □□□□
        □□①□
        □□□□
        --
        □□□□
        □□□□
        □□□□
        □□□①
        --
        """
    expected = [
        TapNote(BeatsTime("0.00"), NotePosition(0, 0)),
        TapNote(BeatsTime("2.75"), NotePosition(1, 1)),
        TapNote(BeatsTime("5.50"), NotePosition(2, 2)),
        TapNote(BeatsTime("8.25"), NotePosition(3, 3)),
    ]
    compare_chart_notes(chart, expected)


def test_irregular_beats_per_frame_2() -> None:
    chart = """
        b=1
        ①□□□
        □□□□
        □□□□
        □□□□
        --
        □□□□
        □①□□
        □□□□
        □□□□
        --
        b=2.75
        □□□□
        □□□□
        □□①□
        □□□□
        --
        □□□□
        □□□□
        □□□□
        □□□①
        --
        """
    expected = [
        TapNote(BeatsTime("0.00"), NotePosition(0, 0)),
        TapNote(BeatsTime("1.00"), NotePosition(1, 1)),
        TapNote(BeatsTime("2.00"), NotePosition(2, 2)),
        TapNote(BeatsTime("4.75"), NotePosition(3, 3)),
    ]
    compare_chart_notes(chart, expected)


def test_long_notes() -> None:
    chart = """
        #holdbyarrow=1
        ①□□＜
        □□□□
        □□□□
        □□□□
        --
        ①□□□
        □□□□
        □□□□
        □□□□
        --
        """
    expected = [
        LongNote(
            time=BeatsTime(0),
            position=NotePosition(0, 0),
            duration=BeatsTime(4),
            tail_tip=NotePosition(3, 0),
        )
    ]
    compare_chart_notes(chart, expected)


def test_long_notes_ambiguous_case() -> None:
    chart = """
        #holdbyarrow=1
        ①①＜＜
        □□□□
        □□□□
        □□□□
        --
        ①①□□
        □□□□
        □□□□
        □□□□
        --
        """
    expected = [
        LongNote(BeatsTime(0), NotePosition(x, y), BeatsTime(4), NotePosition(tx, ty))
        for (x, y), (tx, ty) in [
            ((0, 0), (2, 0)),
            ((1, 0), (3, 0)),
        ]
    ]
    with pytest.warns(UserWarning):
        compare_chart_notes(chart, expected)


@pytest.mark.filterwarnings("error")
def test_long_notes_simple_solution_no_warning() -> None:
    chart = """
        #holdbyarrow=1
        □□□□
        ＞①①＜
        □□□□
        □□□□
        --
        □□□□
        □①①□
        □□□□
        □□□□
        --
        """
    expected = [
        LongNote(BeatsTime(0), NotePosition(x, y), BeatsTime(4), NotePosition(tx, ty))
        for (x, y), (tx, ty) in [
            ((1, 1), (0, 1)),
            ((2, 1), (3, 1)),
        ]
    ]
    compare_chart_notes(chart, expected)


def test_long_notes_complex_case() -> None:
    chart = """
        #holdbyarrow=1
        □□□□
        □□∨□
        □∨□□
        ＞①①①
        --
        □□□□
        □□□□
        □□□□
        □①①①
        --
        """
    expected = [
        LongNote(BeatsTime(0), NotePosition(x, y), BeatsTime(4), NotePosition(tx, ty))
        for (x, y), (tx, ty) in [
            ((1, 3), (1, 2)),
            ((2, 3), (2, 1)),
            ((3, 3), (0, 3)),
        ]
    ]
    compare_chart_notes(chart, expected)


def test_circle_free() -> None:
    chart = """
        #holdbyarrow=1
        #circlefree=1
        □□□□
        □□□□
        □□□□
        ＞□□①
        --
        □□□□
        □□□□
        □□□□
        □□□13
        --
        """
    expected = [
        LongNote(BeatsTime(0), NotePosition(3, 3), BeatsTime(7), NotePosition(0, 3))
    ]
    compare_chart_notes(chart, expected)



from decimal import Decimal
from pathlib import Path
from typing import Set, Union

from hypothesis import example, given
from hypothesis import note as hypothesis_note
from hypothesis import strategies as st

from jubeatools import song
from jubeatools.formats.format_names import Format
from jubeatools.formats.jubeat_analyser.memo.dump import _dump_memo_chart
from jubeatools.formats.jubeat_analyser.memo.load import MemoParser
from jubeatools.testutils import strategies as jbst
from jubeatools.testutils.test_patterns import dump_and_load_then_compare

from ..test_utils import memo_compatible_song
from . import example1, example2, example3


@given(jbst.notes())
@example(example1.notes)
def test_that_notes_roundtrip(notes: Set[Union[song.TapNote, song.LongNote]]) -> None:
    timing = song.Timing(
        events=[song.BPMEvent(song.BeatsTime(0), Decimal(120))],
        beat_zero_offset=song.SecondsTime(0),
    )
    chart = song.Chart(
        level=Decimal(0),
        timing=timing,
        notes=sorted(notes, key=lambda n: (n.time, n.position)),
    )
    metadata = song.Metadata("", "", Path(""), Path(""))
    string_io = _dump_memo_chart("", chart, metadata, timing, False)
    chart_text = string_io.getvalue()
    hypothesis_note(f"Chart :\n{chart_text}")
    parser = MemoParser()
    for line in chart_text.split("\n"):
        parser.load_line(line)
    parser.finish_last_few_notes()
    actual = set(parser.notes())
    assert notes == actual


@given(memo_compatible_song(), st.booleans())
@example(*example2.data)
@example(*example3.data)
def test_that_full_chart_roundtrips(song: song.Song, circle_free: bool) -> None:
    dump_and_load_then_compare(
        Format.MEMO,
        song,
        bytes_decoder=lambda b: b.decode("shift-jis-2004", errors="surrogateescape"),
        dump_options={"circle_free": circle_free},
    )



from hypothesis import given
from hypothesis import strategies as st

from ..command import escape_string_value, unescape_string_value


@given(st.text())
def test_that_escaping_preserves_string(expected: str) -> None:
    escaped = escape_string_value(expected)
    actual = unescape_string_value(escaped)
    assert expected == actual



from decimal import Decimal
from pathlib import Path
from typing import List, Set, Union

import hypothesis.strategies as st
from hypothesis import given

from jubeatools.formats import Format
from jubeatools.formats.jubeat_analyser.mono_column.dump import _dump_mono_column_chart
from jubeatools.formats.jubeat_analyser.mono_column.load import MonoColumnParser
from jubeatools.song import (
    BeatsTime,
    BPMEvent,
    Chart,
    LongNote,
    Metadata,
    SecondsTime,
    Song,
    TapNote,
    Timing,
)
from jubeatools.testutils.strategies import long_note
from jubeatools.testutils.strategies import notes as notes_strat
from jubeatools.testutils.strategies import tap_note
from jubeatools.testutils.test_patterns import dump_and_load_then_compare

from ..test_utils import memo_compatible_song


@given(st.sets(tap_note(), min_size=1, max_size=100))
def test_that_a_set_of_tap_notes_roundtrip(notes: Set[TapNote]) -> None:
    timing = Timing(
        events=[BPMEvent(BeatsTime(0), Decimal(120))], beat_zero_offset=SecondsTime(0)
    )
    chart = Chart(
        level=Decimal(0),
        timing=timing,
        notes=sorted(notes, key=lambda n: (n.time, n.position)),
    )
    metadata = Metadata("", "", Path(""), Path(""))
    string_io = _dump_mono_column_chart("", chart, metadata, timing)
    chart_text = string_io.getvalue()
    parser = MonoColumnParser()
    for line in chart_text.split("\n"):
        parser.load_line(line)
    actual = set(parser.notes())
    assert notes == actual


@given(long_note())
def test_that_a_single_long_note_roundtrips(note: LongNote) -> None:
    timing = Timing(
        events=[BPMEvent(BeatsTime(0), Decimal(120))], beat_zero_offset=SecondsTime(0)
    )
    chart = Chart(level=Decimal(0), timing=timing, notes=[note])
    metadata = Metadata("", "", Path(""), Path(""))
    string_io = _dump_mono_column_chart("", chart, metadata, timing)
    chart_text = string_io.getvalue()
    parser = MonoColumnParser()
    for line in chart_text.split("\n"):
        parser.load_line(line)
    actual = set(parser.notes())
    assert set([note]) == actual


@given(notes_strat())
def test_that_many_notes_roundtrip(notes: List[Union[TapNote, LongNote]]) -> None:
    timing = Timing(
        events=[BPMEvent(BeatsTime(0), Decimal(120))], beat_zero_offset=SecondsTime(0)
    )
    chart = Chart(
        level=Decimal(0),
        timing=timing,
        notes=sorted(notes, key=lambda n: (n.time, n.position)),
    )
    metadata = Metadata("", "", Path(""), Path(""))
    string_io = _dump_mono_column_chart("", chart, metadata, timing)
    chart_text = string_io.getvalue()
    parser = MonoColumnParser()
    for line in chart_text.split("\n"):
        parser.load_line(line)
    actual = set(parser.notes())
    assert notes == actual


@given(memo_compatible_song(), st.booleans())
def test_that_full_chart_roundtrips(song: Song, circle_free: bool) -> None:
    dump_and_load_then_compare(
        Format.MONO_COLUMN,
        song,
        bytes_decoder=lambda b: b.decode("shift-jis-2004", errors="surrogateescape"),
        dump_options={"circle_free": circle_free},
    )



import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator

from hypothesis import strategies as st

from jubeatools import song
from jubeatools.testutils import strategies as jbst


@st.composite
def memo_compatible_metadata(draw: st.DrawFn) -> song.Metadata:
    # some ranges that are valid in shift-jis
    text_strat = st.text(
        alphabet=st.one_of(
            *(
                st.characters(min_codepoint=a, max_codepoint=b)
                for a, b in (
                    (0x20, 0x7F),
                    (0xB6, 0x109),
                    (0x410, 0x44F),
                    (0x24D0, 0x24E9),
                    (0x3041, 0x3096),
                    (0x309B, 0x30FF),
                    (0xFA30, 0xFA6A),
                    (0xFF01, 0xFF3B),
                    (0xFF3D, 0xFF5D),
                    (0xFF61, 0xFF9F),
                )
            )
        )
    )
    metadata: song.Metadata = draw(
        jbst.metadata(text_strat=text_strat, path_strat=text_strat)
    )
    metadata.preview = None
    metadata.preview_file = None
    return metadata


@st.composite
def memo_compatible_song(draw: st.DrawFn) -> song.Song:
    """Memo only supports one difficulty per file"""
    diff = draw(st.sampled_from(list(d.value for d in song.Difficulty)))
    chart = draw(
        jbst.chart(
            timing_strat=jbst.timing_info(with_bpm_changes=True),
            notes_strat=jbst.notes(),
        )
    )
    metadata: song.Metadata = draw(memo_compatible_metadata())
    return song.Song(
        metadata=metadata,
        charts={diff: chart},
    )


@contextmanager
def temp_file_named_txt() -> Iterator[Path]:
    with tempfile.NamedTemporaryFile(suffix=".txt") as dst:
        yield Path(dst.name)



from copy import deepcopy
from dataclasses import astuple, dataclass
from decimal import Decimal
from itertools import product
from pathlib import Path
from typing import Any, Dict, Iterator, List, Mapping, Set, Tuple, Union

from more_itertools import mark_ends

from jubeatools.song import (
    BeatsTime,
    Chart,
    LongNote,
    Metadata,
    NotePosition,
    Preview,
    SecondsTime,
    Song,
    TapNote,
    Timing,
)
from jubeatools.utils import none_or

from ..command import is_command, parse_command
from ..load_tools import (
    CIRCLE_FREE_TO_NOTE_SYMBOL,
    EMPTY_BEAT_SYMBOLS,
    DoubleColumnChartLine,
    DoubleColumnFrame,
    JubeatAnalyserParser,
    UnfinishedLongNote,
    find_long_note_candidates,
    is_double_column_chart_line,
    is_empty_line,
    load_folder,
    parse_double_column_chart_line,
    pick_correct_long_note_candidates,
)
from ..symbols import CIRCLE_FREE_SYMBOLS


class Memo1Frame(DoubleColumnFrame):
    @property
    def duration(self) -> BeatsTime:
        # This is wrong for the last frame in a section if the section has a
        # length in beats that's not an integer
        return BeatsTime(len(self.timing_part))


@dataclass
class Memo1LoadedSection:
    frames: List[Memo1Frame]
    length: BeatsTime
    tempo: Decimal

    def __str__(self) -> str:
        res = []
        if self.length != 4:
            res += [f"b={self.length}", ""]
        for _, is_last, frame in mark_ends(self.frames):
            res += [str(frame)]
            if not is_last:
                res += [""]
        return "\n".join(res)


class Memo1Parser(JubeatAnalyserParser):

    FORMAT_TAG = "#memo1"

    def __init__(self) -> None:
        super().__init__()
        self.current_chart_lines: List[DoubleColumnChartLine] = []
        self.current_frames: List[Memo1Frame] = []
        self.sections: List[Memo1LoadedSection] = []

    def do_memo1(self) -> None:
        ...

    def do_b(self, value: str) -> None:
        """Because of the way the parser works,
        b= commands must mark the end of a section to be properly taken into
        account"""
        if self.current_chart_lines:
            raise SyntaxError("Found a b= command before the end of a frame")
        if self.current_frames and self._frames_duration() < self.beats_per_section:
            raise SyntaxError("Found a b= command before the end of a section")

        self._push_section()
        super().do_b(value)

    def do_bpp(self, value: str) -> None:
        if self.sections or self.current_frames:
            raise ValueError("jubeatools does not handle changes of #bpp halfway")
        else:
            self._do_bpp(value)

    def append_chart_line(self, line: DoubleColumnChartLine) -> None:
        line.raise_if_unfit(self.bytes_per_panel)
        self.current_chart_lines.append(line)
        if len(self.current_chart_lines) == 4:
            self._push_frame()

    def _frames_duration(self) -> BeatsTime:
        return sum(
            (frame.duration for frame in self.current_frames), start=BeatsTime(0)
        )

    def _current_beat(self) -> BeatsTime:
        # If we've already seen enough beats, we need to circumvent the wrong
        # duration computation
        if self._frames_duration() >= self.beats_per_section:
            frames_duration = self.beats_per_section
        else:
            frames_duration = self._frames_duration()

        return self.section_starting_beat + frames_duration

    def _push_frame(self) -> None:
        position_part = [
            self._split_chart_line(memo_line.position)
            for memo_line in self.current_chart_lines
        ]
        timing_part = [
            self._split_chart_line(memo_line.timing)
            for memo_line in self.current_chart_lines
            if memo_line.timing is not None
        ]
        frame = Memo1Frame(position_part, timing_part)
        # if the current frame has some timing info
        if frame.duration > 0:
            # and the previous frames already cover enough beats
            if self._frames_duration() >= self.beats_per_section:
                # then the current frame starts a new section
                self._push_section()

        self.current_frames.append(frame)
        self.current_chart_lines = []

    def _push_section(self) -> None:
        """Take all currently stacked frames and push them to a new section,
        Move time forward by the number of beats per section"""
        if not self.current_frames:
            raise RuntimeError(
                "Tried pushing a new section but no frames are currently stacked"
            )
        self.sections.append(
            Memo1LoadedSection(
                frames=deepcopy(self.current_frames),
                length=self.beats_per_section,
                tempo=self.current_tempo,
            )
        )
        self.current_frames = []
        self.section_starting_beat += self.beats_per_section

    def finish_last_few_notes(self) -> None:
        """Call this once when the end of the file is reached,
        flushes the chart line and chart frame buffers to create the last chart
        section"""
        if self.current_chart_lines:
            if len(self.current_chart_lines) != 4:
                raise SyntaxError(
                    f"Unfinished chart frame when flushing : {self.current_chart_lines}"
                )
            self._push_frame()
        self._push_section()

    def load_line(self, raw_line: str) -> None:
        line = raw_line.strip()
        self.raise_if_separator(line, self.FORMAT_TAG)
        if is_command(line):
            command, value = parse_command(line)
            self.handle_command(command, value)
        elif is_empty_line(line) or self.is_short_line(line):
            return
        elif is_double_column_chart_line(line):
            memo_chart_line = parse_double_column_chart_line(line)
            self.append_chart_line(memo_chart_line)
        else:
            raise SyntaxError(f"not a valid memo1 file line : {line}")

    def notes(self) -> Iterator[Union[TapNote, LongNote]]:
        if self.hold_by_arrow:
            yield from self._iter_notes()
        else:
            yield from self._iter_notes_without_longs()

    def _iter_frames(
        self,
    ) -> Iterator[
        Tuple[Mapping[str, BeatsTime], Memo1Frame, BeatsTime, Memo1LoadedSection]
    ]:
        """iterate over tuples of
        currently_defined_symbols, frame, section_starting_beat, section"""
        local_symbols = {}
        section_starting_beat = BeatsTime(0)
        for section in self.sections:
            frame_starting_beat = BeatsTime(0)
            for i, frame in enumerate(section.frames):
                if frame.timing_part:
                    frame_starting_beat = sum(
                        (f.duration for f in section.frames[:i]), start=BeatsTime(0)
                    )
                    local_symbols = {
                        symbol: BeatsTime(symbol_index, len(bar))
                        + bar_index
                        + frame_starting_beat
                        for bar_index, bar in enumerate(frame.timing_part)
                        for symbol_index, symbol in enumerate(bar)
                        if symbol not in EMPTY_BEAT_SYMBOLS
                    }
                yield local_symbols, frame, section_starting_beat, section
            section_starting_beat += section.length

    def _iter_notes(self) -> Iterator[Union[TapNote, LongNote]]:
        unfinished_longs: Dict[NotePosition, UnfinishedLongNote] = {}
        for (
            currently_defined_symbols,
            frame,
            section_starting_beat,
            section,
        ) in self._iter_frames():
            should_skip: Set[NotePosition] = set()

            # 1/3 : look for ends to unfinished long notes
            for pos, unfinished_long in unfinished_longs.items():
                x, y = astuple(pos)
                symbol = frame.position_part[y][x]
                if self.circle_free and symbol in CIRCLE_FREE_SYMBOLS:
                    circled_symbol = CIRCLE_FREE_TO_NOTE_SYMBOL[symbol]
                    try:
                        symbol_time = currently_defined_symbols[circled_symbol]
                    except KeyError:
                        raise SyntaxError(
                            "Chart section positional part constains the circle free "
                            f"symbol '{symbol}' but the associated circled symbol "
                            f"'{circled_symbol}' could not be found in the timing part:\n"
                            f"{section}"
                        )
                else:
                    try:
                        symbol_time = currently_defined_symbols[symbol]
                    except KeyError:
                        continue

                should_skip.add(pos)
                note_time = section_starting_beat + symbol_time
                yield unfinished_long.ends_at(note_time)

            unfinished_longs = {
                k: unfinished_longs[k] for k in unfinished_longs.keys() - should_skip
            }

            # 2/3 : look for new long notes starting on this bloc
            arrow_to_note_candidates = find_long_note_candidates(
                frame.position_part, currently_defined_symbols.keys(), should_skip
            )
            if arrow_to_note_candidates:
                solution = pick_correct_long_note_candidates(
                    arrow_to_note_candidates,
                    frame.position_part,
                )
                for arrow_pos, note_pos in solution.items():
                    should_skip.add(arrow_pos)
                    should_skip.add(note_pos)
                    symbol = frame.position_part[note_pos.y][note_pos.x]
                    symbol_time = currently_defined_symbols[symbol]
                    note_time = section_starting_beat + symbol_time
                    unfinished_longs[note_pos] = UnfinishedLongNote(
                        time=note_time, position=note_pos, tail_tip=arrow_pos
                    )

            # 3/3 : find regular notes
            for y, x in product(range(4), range(4)):
                position = NotePosition(x, y)
                if position in should_skip:
                    continue
                symbol = frame.position_part[y][x]
                try:
                    symbol_time = currently_defined_symbols[symbol]
                except KeyError:
                    continue
                note_time = section_starting_beat + symbol_time
                yield TapNote(note_time, position)

    def _iter_notes_without_longs(self) -> Iterator[TapNote]:
        for (
            currently_defined_symbols,
            frame,
            section_starting_beat,
            _,
        ) in self._iter_frames():
            # cross compare symbols with the position information
            for y, x in product(range(4), range(4)):
                symbol = frame.position_part[y][x]
                try:
                    symbol_time = currently_defined_symbols[symbol]
                except KeyError:
                    continue
                note_time = section_starting_beat + symbol_time
                position = NotePosition(x, y)
                yield TapNote(note_time, position)


def _load_memo1_file(lines: List[str]) -> Song:
    parser = Memo1Parser()
    for i, raw_line in enumerate(lines, start=1):
        try:
            parser.load_line(raw_line)
        except Exception as e:
            raise SyntaxError(f"On line {i}\n{e}")

    parser.finish_last_few_notes()
    metadata = Metadata(
        title=parser.title,
        artist=parser.artist,
        audio=none_or(Path, parser.music),
        cover=none_or(Path, parser.jacket),
    )
    if parser.preview_start is not None:
        metadata.preview = Preview(
            start=SecondsTime(parser.preview_start) / 1000, length=SecondsTime(10)
        )

    timing = Timing(
        events=parser.timing_events, beat_zero_offset=SecondsTime(parser.offset) / 1000
    )
    charts = {
        parser.difficulty
        or "EXT": Chart(
            level=Decimal(parser.level),
            timing=timing,
            notes=sorted(parser.notes(), key=lambda n: (n.time, n.position)),
        )
    }
    return Song(metadata=metadata, charts=charts)


def load_memo1(path: Path, **kwargs: Any) -> Song:
    files = load_folder(path)
    charts = [_load_memo1_file(lines) for _, lines in files.items()]
    return Song.from_monochart_instances(*charts)



from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum
from typing import Any, List, Optional, Tuple, Union

from marshmallow import EXCLUDE, Schema, post_dump
from marshmallow.validate import Range
from marshmallow_dataclass import NewType, class_schema


@dataclass
class SongInfo:
    title: Optional[str]
    artist: Optional[str]
    id: Optional[int]


class Mode(int, Enum):
    KEY = 0  # Vertical Scrolling Rhythm Game
    # 1 : Unused
    # 2 : Unused
    CATCH = 3  # EZ2CATCH / Catch the Beat
    PAD = 4  # Jubeat
    TAIKO = 5  # Taiko no Tatsujin
    RING = 6  # Reminds me of Beatstream ?


@dataclass
class Metadata:
    cover: Optional[str]  # path to album art ?
    creator: Optional[str]  # Chart author
    background: Optional[str]  # path to background image
    version: Optional[str]  # freeform difficulty name
    id: Optional[int]
    mode: int
    time: Optional[int]  # creation timestamp ?
    song: SongInfo


PositiveInt = NewType("PositiveInt", int, validate=Range(min=0))
BeatTime = Tuple[PositiveInt, PositiveInt, PositiveInt]
StrictlyPositiveDecimal = NewType(
    "StrictlyPositiveDecimal", Decimal, validate=Range(min=0, min_inclusive=False)
)


@dataclass
class BPMEvent:
    beat: BeatTime
    bpm: StrictlyPositiveDecimal


ButtonIndex = NewType("ButtonIndex", int, validate=Range(min=0, max=15))


@dataclass
class TapNote:
    beat: BeatTime
    index: ButtonIndex


@dataclass
class LongNote:
    beat: BeatTime
    index: ButtonIndex
    endbeat: BeatTime
    endindex: ButtonIndex


@dataclass
class Sound:
    """Used both for the background music and keysounds"""

    beat: BeatTime
    sound: str  # audio file path
    type: int
    offset: int
    isBgm: Optional[bool]
    vol: Optional[int]  # Volume, out of 100
    x: Optional[int]


# TODO: find a keysounded chart to discovery the other values
class SoundType(int, Enum):
    BACKGROUND_MUSIC = 1


Event = Union[Sound, LongNote, TapNote]


@dataclass
class Chart:
    meta: Metadata
    time: List[BPMEvent] = field(default_factory=list)
    note: List[Event] = field(default_factory=list)


class BaseSchema(Schema):
    class Meta:
        ordered = True
        unknown = EXCLUDE

    @post_dump
    def remove_none_values(self, data: dict, **kwargs: Any) -> dict:
        return {key: value for key, value in data.items() if value is not None}


CHART_SCHEMA = class_schema(Chart, base_schema=BaseSchema)()



from collections import defaultdict
from dataclasses import dataclass, field
from fractions import Fraction
from io import StringIO
from itertools import zip_longest
from math import ceil
from typing import Dict, Iterator, List, Union

from more_itertools import collapse, intersperse, mark_ends, windowed

from jubeatools.info import __repository_url__, __version__
from jubeatools.song import (
    BeatsTime,
    Chart,
    LongNote,
    Metadata,
    NotePosition,
    TapNote,
    Timing,
)
from jubeatools.utils import lcm

from ..dump_tools import (
    DIRECTION_TO_ARROW,
    DIRECTION_TO_LINE,
    NOTE_TO_CIRCLE_FREE_SYMBOL,
    JubeatAnalyserDumpedSection,
    LongNoteEnd,
    create_sections_from_chart,
    make_full_dumper_from_jubeat_analyser_chart_dumper,
)
from ..symbols import NOTE_SYMBOLS

AnyNote = Union[TapNote, LongNote, LongNoteEnd]

EMPTY_BEAT_SYMBOL = "−"  # U+02212 : MINUS SIGN
EMPTY_POSITION_SYMBOL = "□"  # U+025A1 : WHITE SQUARE


@dataclass
class Frame:
    positions: Dict[NotePosition, str] = field(default_factory=dict)
    bars: Dict[int, List[str]] = field(default_factory=dict)

    def dump(self, length: BeatsTime) -> Iterator[str]:
        # Check that bars are contiguous
        for a, b in windowed(sorted(self.bars), 2):
            if a is not None and b is not None:
                if b - a != 1:
                    raise ValueError("Frame has discontinuous bars")
        # Check all bars are in the same 4-bar group
        if self.bars.keys() != set(bar % 4 for bar in self.bars):
            raise ValueError("Frame contains bars from different 4-bar groups")

        for pos, bar in zip_longest(self.dump_positions(), self.dump_bars(length)):
            if bar is None:
                bar = ""
            yield f"{pos} {bar}"

    def dump_positions(self) -> Iterator[str]:
        for y in range(4):
            yield "".join(
                self.positions.get(NotePosition(x, y), EMPTY_POSITION_SYMBOL)
                for x in range(4)
            )

    def dump_bars(self, length: BeatsTime) -> Iterator[str]:
        for i in range(ceil(length)):
            if i in self.bars:
                yield f"|{''.join(self.bars[i])}|"
            else:
                yield ""


class Memo1DumpedSection(JubeatAnalyserDumpedSection):
    def render(self, circle_free: bool = False) -> str:
        blocs = []
        commands = list(self._dump_commands())
        if commands:
            blocs.append(commands)
        notes = list(self._dump_notes(circle_free))
        if notes:
            blocs.append(notes)
        return "\n".join(collapse(intersperse("", blocs)))

    def _dump_notes(self, circle_free: bool = False) -> Iterator[str]:
        # Split notes into bars
        notes_by_bar: Dict[int, List[AnyNote]] = defaultdict(list)
        for note in self.notes:
            time_in_section = note.time - self.current_beat
            bar_index = int(time_in_section)
            notes_by_bar[bar_index].append(note)

        # Pre-render timing bars
        bars: Dict[int, List[str]] = defaultdict(list)
        chosen_symbols: Dict[BeatsTime, str] = {}
        symbols_iterator = iter(NOTE_SYMBOLS)
        for bar_index in range(ceil(self.length)):
            notes = notes_by_bar.get(bar_index, [])
            bar_length = lcm(
                *((note.time - self.current_beat).denominator for note in notes)
            )
            if bar_length < 3:
                bar_length = 4
            bar_dict: Dict[int, str] = {}
            for note in notes:
                time_in_section = note.time - self.current_beat
                time_in_bar = time_in_section % Fraction(1)
                time_index = time_in_bar.numerator * (
                    bar_length // time_in_bar.denominator
                )
                if time_index not in bar_dict:
                    symbol = next(symbols_iterator)
                    chosen_symbols[time_in_section] = symbol
                    bar_dict[time_index] = symbol
            bar = [bar_dict.get(i, EMPTY_BEAT_SYMBOL) for i in range(bar_length)]
            bars[bar_index] = bar

        # Create frame by bar
        frames_by_bar: Dict[int, List[Frame]] = defaultdict(list)
        for bar_index in range(ceil(self.length)):
            bar = bars.get(bar_index, [])
            frame = Frame()
            frame.bars[bar_index] = bar
            for note in notes_by_bar[bar_index]:
                time_in_section = note.time - self.current_beat
                symbol = chosen_symbols[time_in_section]
                if isinstance(note, TapNote):
                    if note.position in frame.positions:
                        frames_by_bar[bar_index].append(frame)
                        frame = Frame()
                    frame.positions[note.position] = symbol
                elif isinstance(note, LongNote):
                    needed_positions = set(note.positions_covered())
                    if needed_positions & frame.positions.keys():
                        frames_by_bar[bar_index].append(frame)
                        frame = Frame()
                    direction = note.tail_direction()
                    arrow = DIRECTION_TO_ARROW[direction]
                    line = DIRECTION_TO_LINE[direction]
                    for is_first, is_last, pos in mark_ends(note.positions_covered()):
                        if is_first:
                            frame.positions[pos] = symbol
                        elif is_last:
                            frame.positions[pos] = arrow
                        else:
                            frame.positions[pos] = line
                elif isinstance(note, LongNoteEnd):
                    if note.position in frame.positions:
                        frames_by_bar[bar_index].append(frame)
                        frame = Frame()
                    if circle_free and symbol in NOTE_TO_CIRCLE_FREE_SYMBOL:
                        symbol = NOTE_TO_CIRCLE_FREE_SYMBOL[symbol]
                    frame.positions[note.position] = symbol
            frames_by_bar[bar_index].append(frame)

        # Merge bar-specific frames is possible
        final_frames: List[Frame] = []
        for bar_index in range(ceil(self.length)):
            frames = frames_by_bar[bar_index]
            # Merge if :
            #  - No split in current bar (only one frame)
            #  - There is a previous frame
            #  - The previous frame is not a split frame (it holds a bar)
            #  - The previous and current bars are all in the same 4-bar group
            #  - The note positions in the previous frame do not clash with the current frame
            if (
                len(frames) == 1
                and final_frames
                and final_frames[-1].bars
                and max(final_frames[-1].bars.keys()) // 4
                == min(frames[0].bars.keys()) // 4
                and (
                    not (final_frames[-1].positions.keys() & frames[0].positions.keys())
                )
            ):
                final_frames[-1].bars.update(frames[0].bars)
                final_frames[-1].positions.update(frames[0].positions)
            else:
                final_frames.extend(frames)

        dumped_frames = map(lambda f: f.dump(self.length), final_frames)
        yield from collapse(intersperse("", dumped_frames))


def _raise_if_unfit_for_memo1(
    chart: Chart, timing: Timing, circle_free: bool = False
) -> None:
    if len(timing.events) < 1:
        raise ValueError("No BPM found in file") from None

    first_bpm = min(timing.events, key=lambda e: e.time)
    if first_bpm.time != 0:
        raise ValueError("First BPM event does not happen on beat zero")

    if any(
        not note.has_straight_tail()
        for note in chart.notes
        if isinstance(note, LongNote)
    ):
        raise ValueError(
            "Chart contains diagonal long notes, reprensenting these in"
            " memo format is not supported by jubeatools"
        )


def _section_factory(b: BeatsTime) -> Memo1DumpedSection:
    return Memo1DumpedSection(current_beat=b)


def _dump_memo1_chart(
    difficulty: str,
    chart: Chart,
    metadata: Metadata,
    timing: Timing,
    circle_free: bool = False,
) -> StringIO:

    _raise_if_unfit_for_memo1(chart, timing, circle_free)

    sections = create_sections_from_chart(
        Memo1DumpedSection, chart, difficulty, timing, metadata, circle_free
    )

    # Jubeat Analyser format command
    sections[BeatsTime(0)].commands["memo1"] = None

    # Actual output to file
    file = StringIO()
    file.write(f"// Converted using jubeatools {__version__}\n")
    file.write(f"// {__repository_url__}\n\n")
    file.write(
        "\n\n".join(section.render(circle_free) for _, section in sections.items())
    )

    return file


dump_memo1 = make_full_dumper_from_jubeat_analyser_chart_dumper(_dump_memo1_chart)



import json
import re
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Type

from .format_names import Format


def guess_format(path: Path) -> Format:
    """Try to guess the format of the given file, raise an exception if the
    format is unknown"""
    if path.is_dir():
        raise ValueError("Can't guess chart format for a folder")

    try:
        return recognize_json_formats(path)
    except (json.JSONDecodeError, UnicodeDecodeError, ValueError):
        pass

    try:
        return recognize_jubeat_analyser_format(path)
    except (UnicodeDecodeError, ValueError):
        pass

    if looks_like_eve(path):
        return Format.EVE

    if looks_like_jbsq(path):
        return Format.JBSQ

    raise ValueError("Unrecognized file format")


def recognize_json_formats(path: Path) -> Format:
    with path.open() as f:
        obj = json.load(f)

    if not isinstance(obj, dict):
        raise ValueError("Top level value is not an object")

    if obj.keys() & {"metadata", "data", "version"}:
        return recognize_memon_version(obj)
    elif obj.keys() >= {"meta", "time", "note"}:
        return Format.MALODY
    else:
        raise ValueError("Unrecognized file format")


def recognize_memon_version(obj: dict) -> Format:
    try:
        version = obj["version"]
    except KeyError:
        return Format.MEMON_LEGACY

    if version == "0.1.0":
        return Format.MEMON_0_1_0
    elif version == "0.2.0":
        return Format.MEMON_0_2_0
    elif version == "0.3.0":
        return Format.MEMON_0_3_0
    elif version == "1.0.0":
        return Format.MEMON_1_0_0
    else:
        raise ValueError(f"Unsupported memon version : {version}")


JUBEAT_ANALYSER_COMMANDS = {
    "b",
    "m",
    "o",
    "r",
    "t",
    "#lev",
    "#dif",
    "#title",
    "#artist",
}

COMMENT = re.compile(r"//.*")


def _dirty_jba_line_strip(line: str) -> str:
    """This does not deal with '//' in quotes properly,
    thankfully we don't care when looking for an argument-less command"""
    return COMMENT.sub("", line).strip()


def recognize_jubeat_analyser_format(path: Path) -> Format:
    with path.open(encoding="shift-jis-2004", errors="surrogateescape") as f:
        lines = f.readlines()

    saw_jubeat_analyser_commands = False
    for raw_line in lines:
        line = _dirty_jba_line_strip(raw_line)
        if line in ("#memo2", "#boogie"):
            return Format.MEMO_2
        elif line == "#memo1":
            return Format.MEMO_1
        elif line == "#memo":
            return Format.MEMO
        elif "=" in line:
            index = line.index("=")
            if line[:index] in JUBEAT_ANALYSER_COMMANDS:
                saw_jubeat_analyser_commands = True

    if saw_jubeat_analyser_commands:
        return Format.MONO_COLUMN
    else:
        raise ValueError("Unrecognized file format")


def false_if_raises(
    *exceptions: Type[Exception],
) -> Callable[[Callable[..., bool]], Callable[..., bool]]:
    def decorator(f: Callable[..., bool]) -> Callable[..., bool]:
        @wraps(f)
        def wrapper(*a: Any, **kw: Any) -> bool:
            try:
                return f(*a, **kw)
            except Exception as e:
                if exceptions and not isinstance(e, exceptions):
                    raise
                else:
                    return False

        return wrapper

    return decorator


@false_if_raises(UnicodeDecodeError, StopIteration)
def looks_like_eve(path: Path) -> bool:
    with path.open(encoding="ascii") as f:
        return looks_like_eve_line(f.readline())


EVE_COMMANDS = {
    "END",
    "MEASURE",
    "HAKU",
    "PLAY",
    "LONG",
    "TEMPO",
}


def looks_like_eve_line(line: str) -> bool:
    columns = line.split(",")
    if len(columns) != 3:
        return False

    raw_tick, raw_command, raw_value = map(str.strip, columns)
    try:
        int(raw_tick)
    except Exception:
        return False

    if raw_command not in EVE_COMMANDS:
        return False

    try:
        int(raw_value)
    except Exception:
        return False

    return True


def looks_like_jbsq(path: Path) -> bool:
    with path.open(mode="rb") as f:
        magic = f.read(4)
        return magic in (b"IJBQ", b"IJSQ", b"JBSQ")



from copy import deepcopy
from io import StringIO
from typing import Dict, Iterator, List

from more_itertools import collapse, intersperse, mark_ends

from jubeatools.info import __repository_url__, __version__
from jubeatools.song import (
    BeatsTime,
    Chart,
    LongNote,
    Metadata,
    NotePosition,
    TapNote,
    Timing,
)

from ..dump_tools import (
    BEATS_TIME_TO_SYMBOL,
    DEFAULT_EXTRA_SYMBOLS,
    DIRECTION_TO_ARROW,
    DIRECTION_TO_LINE,
    NOTE_TO_CIRCLE_FREE_SYMBOL,
    JubeatAnalyserDumpedSection,
    LongNoteEnd,
    create_sections_from_chart,
    make_full_dumper_from_jubeat_analyser_chart_dumper,
)


class MonoColumnDumpedSection(JubeatAnalyserDumpedSection):
    def render(self, circle_free: bool = False) -> str:
        blocs = []
        commands = list(self._dump_commands())
        if commands:
            blocs.append(commands)
        symbols = list(self._dump_symbol_definitions())
        if symbols:
            blocs.append(symbols)
        notes = list(self._dump_notes(circle_free))
        if notes:
            blocs.append(notes)
        return "\n".join(collapse([intersperse("", blocs), "--"]))

    def _dump_notes(
        self,
        circle_free: bool = False,
    ) -> Iterator[str]:
        frames: List[Dict[NotePosition, str]] = []
        frame: Dict[NotePosition, str] = {}
        for note in self.notes:
            time_in_section = note.time - self.current_beat
            symbol = self.symbols[time_in_section]
            if isinstance(note, LongNote):
                needed_positions = set(note.positions_covered())
                if needed_positions & frame.keys():
                    frames.append(frame)
                    frame = {}
                direction = note.tail_direction()
                arrow = DIRECTION_TO_ARROW[direction]
                line = DIRECTION_TO_LINE[direction]
                for is_first, is_last, pos in mark_ends(note.positions_covered()):
                    if is_first:
                        frame[pos] = symbol
                    elif is_last:
                        frame[pos] = arrow
                    else:
                        frame[pos] = line
            elif isinstance(note, TapNote):
                if note.position in frame:
                    frames.append(frame)
                    frame = {}
                frame[note.position] = symbol
            elif isinstance(note, LongNoteEnd):
                if note.position in frame:
                    frames.append(frame)
                    frame = {}
                if circle_free and symbol in NOTE_TO_CIRCLE_FREE_SYMBOL:
                    symbol = NOTE_TO_CIRCLE_FREE_SYMBOL[symbol]
                frame[note.position] = symbol

        frames.append(frame)
        dumped_frames = map(self._dump_frame, frames)
        yield from collapse(intersperse("", dumped_frames))

    @staticmethod
    def _dump_frame(frame: Dict[NotePosition, str]) -> Iterator[str]:
        for y in range(4):
            yield "".join(frame.get(NotePosition(x, y), "□") for x in range(4))


def _raise_if_unfit_for_mono_column(chart: Chart, timing: Timing) -> None:
    if len(timing.events) < 1:
        raise ValueError("No BPM found in file") from None

    first_bpm = min(timing.events, key=lambda e: e.time)
    if first_bpm.time != 0:
        raise ValueError("First BPM event does not happen on beat zero")

    if any(
        not note.has_straight_tail()
        for note in chart.notes
        if isinstance(note, LongNote)
    ):
        raise ValueError(
            "Chart contains diagonal long notes, reprensenting these in"
            " mono_column format is not supported by jubeatools"
        )


def _section_factory(b: BeatsTime) -> MonoColumnDumpedSection:
    return MonoColumnDumpedSection(current_beat=b)


def _dump_mono_column_chart(
    difficulty: str,
    chart: Chart,
    metadata: Metadata,
    timing: Timing,
    circle_free: bool = False,
) -> StringIO:

    _raise_if_unfit_for_mono_column(chart, timing)

    sections = create_sections_from_chart(
        _section_factory, chart, difficulty, timing, metadata, circle_free
    )

    # Define extra symbols
    existing_symbols = deepcopy(BEATS_TIME_TO_SYMBOL)
    extra_symbols = iter(DEFAULT_EXTRA_SYMBOLS)
    for section_start, section in sections.items():
        # intentionally not a copy : at the end of this loop every section
        # holds a reference to a dict containing every defined symbol
        section.symbols = existing_symbols
        for note in section.notes:
            time_in_section = note.time - section_start
            if time_in_section not in existing_symbols:
                new_symbol = next(extra_symbols)
                section.symbol_definitions[time_in_section] = new_symbol
                existing_symbols[time_in_section] = new_symbol

    # Actual output to file
    file = StringIO()
    file.write(f"// Converted using jubeatools {__version__}\n")
    file.write(f"// {__repository_url__}\n\n")
    for _, section in sections.items():
        file.write(section.render(circle_free) + "\n")

    return file


dump_mono_column = make_full_dumper_from_jubeat_analyser_chart_dumper(
    _dump_mono_column_chart
)



from io import StringIO
from itertools import chain
from typing import Any, Dict, List, Union

import simplejson as json

from jubeatools import song as jbt
from jubeatools.formats.filetypes import SongFile
from jubeatools.utils import lcm

from ..tools import make_memon_dumper
from . import schema


def _long_note_tail_value_v0(note: jbt.LongNote) -> int:
    dx = note.tail_tip.x - note.position.x
    dy = note.tail_tip.y - note.position.y
    try:
        return schema.X_Y_OFFSET_TO_P_VALUE[dx, dy]
    except KeyError:
        raise ValueError(
            f"memon cannot represent a long note with its tail starting ({dx}, {dy}) away from the note"
        ) from None


def _get_timing(song: jbt.Song) -> jbt.Timing:
    if song.common_timing is not None:
        return song.common_timing
    else:
        return next(
            chart.timing for chart in song.charts.values() if chart.timing is not None
        )


def _raise_if_unfit_for_v0(song: jbt.Song, version: str) -> None:
    """Raises an exception if the Song object is ill-formed or contains information
    that cannot be represented in a memon v0.x.y file (includes legacy)"""

    if song.common_timing is None and all(
        chart.timing is None for chart in song.charts.values()
    ):
        raise ValueError("The song has no timing information")

    chart_timings = [
        chart.timing for chart in song.charts.values() if chart.timing is not None
    ]

    if chart_timings:
        first_one = chart_timings[0]
        if any(t != first_one for t in chart_timings):
            raise ValueError(
                f"memon:{version} cannot represent a song with per-chart timing"
            )

    timing = _get_timing(song)
    number_of_timing_events = len(timing.events)
    if number_of_timing_events != 1:
        if number_of_timing_events == 0:
            raise ValueError("The song has no BPM")
        else:
            raise ValueError(f"memon:{version} does not handle BPM changes")

    event = timing.events[0]
    if event.BPM <= 0:
        raise ValueError(f"memon:{version} only accepts strictly positive BPMs")

    if event.time != 0:
        raise ValueError(f"memon:{version} only accepts a BPM on the first beat")

    for difficulty, chart in song.charts.items():
        if len(set(chart.notes)) != len(chart.notes):
            raise ValueError(
                f"{difficulty} chart has duplicate notes, these cannot be represented"
            )


def _dump_to_json(memon: dict) -> bytes:
    memon_fp = StringIO()
    json.dump(memon, memon_fp, use_decimal=True, indent=4)
    return memon_fp.getvalue().encode("utf-8")


def _compute_resolution(notes: List[Union[jbt.TapNote, jbt.LongNote]]) -> int:
    return lcm(
        *chain(
            iter(note.time.denominator for note in notes),
            iter(
                note.duration.denominator
                for note in notes
                if isinstance(note, jbt.LongNote)
            ),
        )
    )


def _dump_memon_note_v0(
    note: Union[jbt.TapNote, jbt.LongNote], resolution: int
) -> Dict[str, int]:
    """converts a note into the {n, t, l, p} form"""
    memon_note = {
        "n": note.position.index,
        "t": note.time.numerator * (resolution // note.time.denominator),
        "l": 0,
        "p": 0,
    }
    if isinstance(note, jbt.LongNote):
        memon_note["l"] = note.duration.numerator * (
            resolution // note.duration.denominator
        )
        memon_note["p"] = _long_note_tail_value_v0(note)

    return memon_note


def _dump_memon_legacy(song: jbt.Song, **kwargs: Any) -> SongFile:
    _raise_if_unfit_for_v0(song, "legacy")
    timing = _get_timing(song)

    memon: Dict[str, Any] = {
        "metadata": {
            "song title": song.metadata.title,
            "artist": song.metadata.artist,
            "music path": str(song.metadata.audio),
            "jacket path": str(song.metadata.cover),
            "BPM": timing.events[0].BPM,
            "offset": -timing.beat_zero_offset,
        },
        "data": [],
    }
    for difficulty, chart in song.charts.items():
        resolution = _compute_resolution(chart.notes)
        memon["data"].append(
            {
                "dif_name": difficulty,
                "level": chart.level,
                "resolution": resolution,
                "notes": [
                    _dump_memon_note_v0(note, resolution)
                    for note in sorted(
                        set(chart.notes), key=lambda n: (n.time, n.position)
                    )
                ],
            }
        )

    return SongFile(contents=_dump_to_json(memon), song=song)


dump_memon_legacy = make_memon_dumper(_dump_memon_legacy)


def _dump_memon_0_1_0(song: jbt.Song, **kwargs: Any) -> SongFile:
    _raise_if_unfit_for_v0(song, "v0.1.0")
    timing = _get_timing(song)

    memon: Dict[str, Any] = {
        "version": "0.1.0",
        "metadata": {
            "song title": song.metadata.title,
            "artist": song.metadata.artist,
            "music path": str(song.metadata.audio),
            "album cover path": str(song.metadata.cover),
            "BPM": timing.events[0].BPM,
            "offset": -timing.beat_zero_offset,
        },
        "data": dict(),
    }
    for difficulty, chart in song.charts.items():
        resolution = _compute_resolution(chart.notes)
        memon["data"][difficulty] = {
            "level": chart.level,
            "resolution": resolution,
            "notes": [
                _dump_memon_note_v0(note, resolution)
                for note in sorted(set(chart.notes), key=lambda n: (n.time, n.position))
            ],
        }

    return SongFile(contents=_dump_to_json(memon), song=song)


dump_memon_0_1_0 = make_memon_dumper(_dump_memon_0_1_0)


def _dump_memon_0_2_0(song: jbt.Song, **kwargs: Any) -> SongFile:
    _raise_if_unfit_for_v0(song, "v0.2.0")
    timing = _get_timing(song)

    memon: Dict[str, Any] = {
        "version": "0.2.0",
        "metadata": {
            "song title": song.metadata.title,
            "artist": song.metadata.artist,
            "music path": str(song.metadata.audio),
            "album cover path": str(song.metadata.cover),
            "BPM": timing.events[0].BPM,
            "offset": -timing.beat_zero_offset,
        },
        "data": {},
    }

    if song.metadata.preview is not None:
        memon["metadata"]["preview"] = {
            "position": song.metadata.preview.start,
            "length": song.metadata.preview.length,
        }

    for difficulty, chart in song.charts.items():
        resolution = _compute_resolution(chart.notes)
        memon["data"][difficulty] = {
            "level": chart.level,
            "resolution": resolution,
            "notes": [
                _dump_memon_note_v0(note, resolution)
                for note in sorted(set(chart.notes), key=lambda n: (n.time, n.position))
            ],
        }

    return SongFile(contents=_dump_to_json(memon), song=song)


dump_memon_0_2_0 = make_memon_dumper(_dump_memon_0_2_0)


def _dump_memon_0_3_0(song: jbt.Song, **kwargs: Any) -> SongFile:
    _raise_if_unfit_for_v0(song, "v0.3.0")
    timing = _get_timing(song)

    memon: Dict[str, Any] = {
        "version": "0.3.0",
        "metadata": {
            "song title": song.metadata.title,
            "artist": song.metadata.artist,
            "BPM": timing.events[0].BPM,
            "offset": -timing.beat_zero_offset,
        },
        "data": {},
    }

    if song.metadata.audio is not None:
        memon["metadata"]["music path"] = str(song.metadata.audio)

    if song.metadata.cover is not None:
        memon["metadata"]["album cover path"] = str(song.metadata.cover)

    if song.metadata.preview is not None:
        memon["metadata"]["preview"] = {
            "position": song.metadata.preview.start,
            "length": song.metadata.preview.length,
        }

    if song.metadata.preview_file is not None:
        memon["metadata"]["preview path"] = str(song.metadata.preview_file)

    for difficulty, chart in song.charts.items():
        resolution = _compute_resolution(chart.notes)
        memon["data"][difficulty] = {
            "level": chart.level,
            "resolution": resolution,
            "notes": [
                _dump_memon_note_v0(note, resolution)
                for note in sorted(set(chart.notes), key=lambda n: (n.time, n.position))
            ],
        }

    return SongFile(contents=_dump_to_json(memon), song=song)


dump_memon_0_3_0 = make_memon_dumper(_dump_memon_0_3_0)



import time
from functools import singledispatch
from pathlib import Path
from typing import List, Optional, Tuple, Union

import simplejson as json

from jubeatools import song
from jubeatools.formats.dump_tools import make_dumper_from_chart_file_dumper
from jubeatools.formats.filetypes import ChartFile
from jubeatools.utils import none_or

from . import schema as malody


def dump_malody_song(song: song.Song, **kwargs: dict) -> List[ChartFile]:
    res = []
    for dif, chart, timing in song.iter_charts_with_applicable_timing():
        malody_chart = dump_malody_chart(song.metadata, dif, chart, timing)
        json_chart = malody.CHART_SCHEMA.dump(malody_chart)
        chart_bytes = json.dumps(json_chart, indent=4, use_decimal=True).encode("utf-8")
        res.append(ChartFile(chart_bytes, song, dif, chart))

    return res


dump_malody = make_dumper_from_chart_file_dumper(
    internal_dumper=dump_malody_song, file_name_template=Path("{difficulty:l}.mc")
)


def dump_malody_chart(
    metadata: song.Metadata, dif: Optional[str], chart: song.Chart, timing: song.Timing
) -> malody.Chart:
    meta = dump_metadata(metadata, dif)
    time = dump_timing(timing)
    notes = dump_notes(chart.notes)
    if metadata.audio is not None:
        notes += [dump_bgm(metadata.audio, timing)]
    return malody.Chart(meta=meta, time=time, note=notes)


def dump_metadata(metadata: song.Metadata, dif: Optional[str]) -> malody.Metadata:
    return malody.Metadata(
        cover=None,
        creator=None,
        background=none_or(str, metadata.cover),
        version=dif,
        id=0,
        mode=malody.Mode.PAD,
        time=int(time.time()),
        song=malody.SongInfo(
            title=metadata.title,
            artist=metadata.artist,
            id=0,
        ),
    )


def dump_timing(timing: song.Timing) -> List[malody.BPMEvent]:
    sorted_events = sorted(timing.events, key=lambda e: e.time)
    return [dump_bpm_change(e) for e in sorted_events]


def dump_bpm_change(b: song.BPMEvent) -> malody.BPMEvent:
    return malody.BPMEvent(
        beat=beats_to_fraction_tuple(b.time),
        bpm=b.BPM,
    )


def dump_notes(notes: List[Union[song.TapNote, song.LongNote]]) -> List[malody.Event]:
    return [dump_note(n) for n in notes]


@singledispatch
def dump_note(
    n: Union[song.TapNote, song.LongNote]
) -> Union[malody.TapNote, malody.LongNote]:
    raise NotImplementedError(f"Unknown note type : {type(n)}")


@dump_note.register
def dump_tap_note(n: song.TapNote) -> malody.TapNote:
    return malody.TapNote(
        beat=beats_to_fraction_tuple(n.time),
        index=n.position.index,
    )


@dump_note.register
def dump_long_note(n: song.LongNote) -> malody.LongNote:
    return malody.LongNote(
        beat=beats_to_fraction_tuple(n.time),
        index=n.position.index,
        endbeat=beats_to_fraction_tuple(n.time + n.duration),
        endindex=n.tail_tip.index,
    )


def dump_bgm(audio: Path, timing: song.Timing) -> malody.Sound:
    return malody.Sound(
        beat=beats_to_fraction_tuple(song.BeatsTime(0)),
        sound=str(audio),
        vol=100,
        offset=-int(timing.beat_zero_offset * 1000),
        type=malody.SoundType.BACKGROUND_MUSIC,
        isBgm=None,
        x=None,
    )


def beats_to_fraction_tuple(b: song.BeatsTime) -> Tuple[int, int, int]:
    integer_part = int(b)
    remainder = b % 1
    return (
        integer_part,
        remainder.numerator,
        remainder.denominator,
    )



from importlib import resources

from jubeatools.formats import LOADERS
from jubeatools.formats.guess import guess_format


def load_test_file(package: resources.Package, file: str) -> None:
    with resources.as_file(resources.files(package) / file) as p:
        format_ = guess_format(p)
        loader = LOADERS[format_]
        _ = loader(p)



from enum import Enum

from hypothesis import given

from jubeatools import song
from jubeatools.formats import Format
from jubeatools.formats.konami.testutils import eve_compatible_song
from jubeatools.testutils.test_patterns import dump_and_load, dump_and_load_then_compare


@given(eve_compatible_song())
def test_that_full_chart_roundtrips(song: song.Song) -> None:
    dump_and_load_then_compare(
        Format.EVE,
        song,
        bytes_decoder=lambda b: b.decode("ascii"),
        load_options={"beat_snap": 12},
    )


@given(eve_compatible_song())
def test_that_difficulty_name_is_loaded_properly(original_song: song.Song) -> None:
    recovered_song = dump_and_load(
        Format.EVE,
        original_song,
        bytes_decoder=lambda b: b.decode("ascii"),
        load_options={"beat_snap": 12},
    )
    original_dif, _ = original_song.charts.popitem()
    recovered_dif, _ = recovered_song.charts.popitem()
    assert type(original_dif) == str
    assert not isinstance(original_dif, Enum)
    assert type(recovered_dif) == str
    assert not isinstance(recovered_dif, Enum)



from typing import Set

import hypothesis.strategies as st
from hypothesis import given

from jubeatools import song
from jubeatools.formats.format_names import Format
from jubeatools.testutils import strategies as jbst
from jubeatools.testutils.test_patterns import dump_and_load_then_compare


@st.composite
def memon_diffs(draw: st.DrawFn) -> Set[str]:
    simple_diff_names = st.sampled_from(list(d.value for d in song.Difficulty))
    diff_names = st.one_of(
        simple_diff_names,
        st.text(
            alphabet=st.characters(min_codepoint=0x20, max_codepoint=0x7E),
            min_size=1,
            max_size=20,
        ),
    )
    s: Set[str] = draw(st.sets(diff_names, min_size=1, max_size=10))
    return s


@st.composite
def memon_legacy_compatible_song(draw: st.DrawFn) -> song.Song:
    """Memon versions below v0.2.0 do not support any preview metadata"""
    random_song: song.Song = draw(
        jbst.song(
            diffs_strat=memon_diffs(),
            chart_strat=jbst.chart(timing_strat=st.none()),
            common_timing_strat=jbst.timing_info(with_bpm_changes=False),
        )
    )
    random_song.metadata.preview = None
    random_song.metadata.preview_file = None
    return random_song


@given(memon_legacy_compatible_song())
def test_memon_legacy(song: song.Song) -> None:
    dump_and_load_then_compare(Format.MEMON_LEGACY, song)


memon_0_1_0_compatible_song = memon_legacy_compatible_song


@given(memon_0_1_0_compatible_song())
def test_memon_0_1_0(song: song.Song) -> None:
    dump_and_load_then_compare(Format.MEMON_0_1_0, song)


@st.composite
def memon_0_2_0_compatible_song(draw: st.DrawFn) -> song.Song:
    """Memon v0.2.0 does not support preview_file"""
    random_song: song.Song = draw(
        jbst.song(
            diffs_strat=memon_diffs(),
            chart_strat=jbst.chart(timing_strat=st.none()),
            common_timing_strat=jbst.timing_info(with_bpm_changes=False),
        )
    )
    random_song.metadata.preview_file = None
    return random_song


@given(memon_0_2_0_compatible_song())
def test_memon_0_2_0(song: song.Song) -> None:
    dump_and_load_then_compare(Format.MEMON_0_2_0, song)


@st.composite
def memon_0_3_0_compatible_song(draw: st.DrawFn) -> song.Song:
    random_song: song.Song = draw(
        jbst.song(
            diffs_strat=memon_diffs(),
            chart_strat=jbst.chart(timing_strat=st.none()),
            common_timing_strat=jbst.timing_info(with_bpm_changes=False),
        )
    )

    return random_song


@given(memon_0_3_0_compatible_song())
def test_memon_0_3_0(song: song.Song) -> None:
    dump_and_load_then_compare(Format.MEMON_0_3_0, song)



"""Malody is a multiplatform rhythm game that mainly lives off content created
by its community, as is common in the rhythm game simulator scene. It support
many different games or "Modes", including jubeat (known as "Pad" Mode)

The file format it uses is not that well documented but is simple enough to
make sense of without docs. It's a json file with some defined schema"""

from .dump import dump_malody
from .load import load_malody



from pathlib import Path
from typing import Set

import hypothesis.strategies as st
from hypothesis import given

from jubeatools import song
from jubeatools.formats.format_names import Format
from jubeatools.testutils import strategies as jbst
from jubeatools.testutils.test_patterns import dump_and_load_then_compare


@st.composite
def memon_diffs(draw: st.DrawFn) -> Set[str]:
    simple_diff_names = st.sampled_from(list(d.value for d in song.Difficulty))
    diff_names = st.one_of(
        simple_diff_names,
        st.text(
            alphabet=st.characters(min_codepoint=0x20, max_codepoint=0x7E),
            min_size=1,
            max_size=20,
        ),
    )
    s: Set[str] = draw(st.sets(diff_names, min_size=1, max_size=10))
    return s


@st.composite
def memon_1_0_0_compatible_song(draw: st.DrawFn) -> song.Song:
    """Memon v1.0.0 only support one kind of metadata at once"""
    random_song: song.Song = draw(
        jbst.song(
            diffs_strat=memon_diffs(),
            common_hakus_strat=st.one_of(st.none(), jbst.hakus()),
            chart_strat=jbst.chart(
                hakus_strat=st.one_of(st.none(), jbst.hakus()),
            ),
        )
    )
    preview = draw(st.one_of(jbst.metadata_path_strat(), jbst.preview()))
    if isinstance(preview, str):
        random_song.metadata.preview = None
        random_song.metadata.preview_file = Path(preview)
    else:
        random_song.metadata.preview = preview
        random_song.metadata.preview_file = None

    return random_song


@given(memon_1_0_0_compatible_song())
def test_memon_1_0_0(song: song.Song) -> None:
    dump_and_load_then_compare(Format.MEMON_1_0_0, song)



"""
This module contains code for the different formats read by "jubeat analyser".

"jubeat analyser" is a Windows program that can play back chart files and
export them to video files, it can also be used as a jubeat simulator.

"memo" is a vague term that refers to several slightly different formats.
My understanding is that they were all originally derived from the (somewhat)
human-readable format choosen by websites like jubeat memo or cosmos memo.
These websites would provide text transcripts of official jubeat charts as
training material for hardcore players.

The machine-readable variants or these text formats are partially documented
(in japanese) on these pages :
- http://yosh52.web.fc2.com/jubeat/fumenformat.html
- http://yosh52.web.fc2.com/jubeat/holdmarker.html
"""

from .memo import dump_memo, load_memo
from .memo1 import dump_memo1, load_memo1
from .memo2 import dump_memo2, load_memo2
from .mono_column import dump_mono_column, load_mono_column



from __future__ import annotations

import math
from dataclasses import astuple, dataclass
from enum import Enum
from fractions import Fraction
from itertools import count
from typing import Iterator, Union

from jubeatools import song
from jubeatools.formats.timemap import TimeMap

AnyNote = Union[song.TapNote, song.LongNote]


DIRECTION_TO_VALUE = {
    song.Direction.DOWN: 0,
    song.Direction.UP: 1,
    song.Direction.RIGHT: 2,
    song.Direction.LEFT: 3,
}

VALUE_TO_DIRECTION = {v: k for k, v in DIRECTION_TO_VALUE.items()}

# int is here to allow sorting
class Command(int, Enum):
    END = 1
    MEASURE = 2
    HAKU = 3
    PLAY = 4
    LONG = 5
    TEMPO = 6


@dataclass(order=True)
class Event:
    """Represents a line in an .eve file or an event struct in a .jbsq file"""

    time: int
    command: Command
    value: int

    def __post_init__(self) -> None:
        try:
            check_func = VALUES_CHECKERS[self.command]
        except KeyError:
            # most likely no check function associated : forget about it
            pass

        try:
            check_func(self.value)
        except ValueError as e:
            raise ValueError(
                f"{self.value} is not a valid value for the {self.command!r} "
                f"command. {e}"
            )

    def dump(self) -> str:
        return f"{self.time:>8},{self.command.name:<8},{self.value:>8}"

    @classmethod
    def from_tap_note(cls, note: song.TapNote, time_map: TimeMap) -> Event:
        ticks = ticks_at_beat(note.time, time_map)
        value = note.position.index
        return Event(time=ticks, command=Command.PLAY, value=value)

    @classmethod
    def from_long_note(cls, note: song.LongNote, time_map: TimeMap) -> Event:
        eve_long = EveLong.from_jubeatools(note, time_map)
        ticks = ticks_at_beat(note.time, time_map)
        return Event(time=ticks, command=Command.LONG, value=eve_long.value)


def is_zero(value: int) -> None:
    if value != 0:
        raise ValueError(f"Value should be zero")


def is_valid_button_index(value: int) -> None:
    # Should raise ValueError if invalid
    _ = song.NotePosition.from_index(value)


def is_valid_tail_position(value: int) -> None:
    # Should raise ValueError if invalid
    _ = EveLong.from_value(value)


def is_not_zero(value: int) -> None:
    if value == 0:
        raise ValueError(f"Value cannot be zero")


VALUES_CHECKERS = {
    Command.END: is_zero,
    Command.MEASURE: is_zero,
    Command.HAKU: is_zero,
    Command.PLAY: is_valid_button_index,
    Command.LONG: is_valid_tail_position,
    Command.TEMPO: is_not_zero,
}


@dataclass
class EveLong:
    duration: int
    length: int
    direction: int
    position: int

    def __post_init__(self) -> None:
        if self.duration < 0:
            raise ValueError("Duration can't be negative")
        if not 1 <= self.length < 4:
            raise ValueError("Tail length must be between 1 and 3 inclusive")
        if not 0 <= self.position < 16:
            raise ValueError("Note Position must be between 0 and 15 inclusive")
        if not 0 <= self.direction < 4:
            raise ValueError("direction value must be between 0 and 3 inclusive")

        pos = song.NotePosition.from_index(self.position)
        direction = VALUE_TO_DIRECTION[self.direction]
        step_vector = song.TAIL_DIRECTION_TO_OUTWARDS_VECTOR[direction]
        tail_pos = pos + (self.length * step_vector)
        if not ((0 <= tail_pos.x < 4) and (0 <= tail_pos.y < 4)):
            raise ValueError(
                f"Long note tail starts on {astuple(tail_pos)} which is "
                "outside the screen"
            )

    @classmethod
    def from_jubeatools(cls, note: song.LongNote, time_map: TimeMap) -> EveLong:
        if not note.has_straight_tail():
            raise ValueError("Diagonal tails cannot be represented in eve format")

        return cls(
            duration=duration_in_ticks(note, time_map),
            length=len(list(note.positions_covered())) - 1,
            direction=DIRECTION_TO_VALUE[note.tail_direction()],
            position=note.position.index,
        )

    @classmethod
    def from_value(cls, value: int) -> EveLong:
        if value < 0:
            raise ValueError("Value cannot be negative")

        position = value & 0b1111  # first 4 bits
        direction = (value >> 4) & 0b11  # next 2 bits
        length = (value >> 6) & 0b11  # next 2 bits
        duration = value >> 8  # remaining bits
        return cls(duration, length, direction, position)

    @property
    def value(self) -> int:
        return (
            (self.duration << 8)
            + (self.length << 6)
            + (self.direction << 4)
            + self.position
        )


def ticks_at_beat(time: song.BeatsTime, time_map: TimeMap) -> int:
    seconds_time = time_map.fractional_seconds_at(time)
    return seconds_to_ticks(seconds_time)


def duration_in_ticks(long: song.LongNote, time_map: TimeMap) -> int:
    press_time = time_map.fractional_seconds_at(long.time)
    release_time = time_map.fractional_seconds_at(long.time + long.duration)
    length_in_seconds = release_time - press_time
    return seconds_to_ticks(length_in_seconds)


def ticks_to_seconds(tick: int) -> Fraction:
    """Convert eve ticks (300 Hz) to seconds"""
    return Fraction(tick, 300)


def seconds_to_ticks(time: Fraction) -> int:
    """Convert fractional seconds to eve ticks (300 Hz)"""
    return round(time * 300)


def value_to_truncated_bpm(value: int) -> Fraction:
    """Only keeps enough significant digits to allow recovering the original
    TEMPO line value from the bpm"""
    exact_bpm = value_to_bpm(value)
    truncated_bpms = iter_truncated(exact_bpm)
    bpms_preserving_value = filter(
        lambda b: bpm_to_value(b) < value + 1, truncated_bpms
    )
    return next(bpms_preserving_value)


def iter_truncated(f: Fraction) -> Iterator[Fraction]:
    for places in count():
        yield truncate_fraction(f, places)


def truncate_fraction(f: Fraction, places: int) -> Fraction:
    """Truncates a fraction to the given number of decimal places"""
    exponent = Fraction(10) ** places
    return Fraction(math.floor(f * exponent), exponent)


def value_to_bpm(value: int) -> Fraction:
    return 6 * 10**7 / Fraction(value)


def bpm_to_value(bpm: Fraction) -> Fraction:
    return 6 * 10**7 / bpm



from decimal import Decimal
from typing import Iterable, List, Optional, Set

from more_itertools import numeric_range

from jubeatools import song
from jubeatools.formats.load_tools import round_beats
from jubeatools.formats.timemap import BPMAtSecond, TimeMap
from jubeatools.utils import group_by

from .commons import (
    VALUE_TO_DIRECTION,
    AnyNote,
    Command,
    EveLong,
    Event,
    ticks_to_seconds,
    value_to_truncated_bpm,
)


def make_chart_from_events(events: Iterable[Event], beat_snap: int = 240) -> song.Chart:
    events_by_command = group_by(events, lambda e: e.command)
    bpms = [
        BPMAtSecond(
            seconds=ticks_to_seconds(e.time), BPM=value_to_truncated_bpm(e.value)
        )
        for e in sorted(events_by_command[Command.TEMPO])
    ]
    time_map = TimeMap.from_seconds(bpms)
    tap_notes: List[AnyNote] = [
        make_tap_note(e.time, e.value, time_map, beat_snap)
        for e in events_by_command[Command.PLAY]
    ]
    long_notes: List[AnyNote] = [
        make_long_note(e.time, e.value, time_map, beat_snap)
        for e in events_by_command[Command.LONG]
    ]
    all_notes = sorted(tap_notes + long_notes, key=lambda n: (n.time, n.position))
    timing = time_map.convert_to_timing_info(beat_snap=beat_snap)
    end_tick = events_by_command[Command.END].pop().time
    hakus = make_hakus(
        [e.time for e in events_by_command[Command.HAKU]],
        end_tick,
        time_map,
        beat_snap,
    )
    return song.Chart(level=Decimal(0), timing=timing, notes=all_notes, hakus=hakus)


def make_tap_note(
    ticks: int, value: int, time_map: TimeMap, beat_snap: int
) -> song.TapNote:
    time = beats_at_tick(ticks, time_map, beat_snap)
    position = song.NotePosition.from_index(value)
    return song.TapNote(time=time, position=position)


def make_long_note(
    ticks: int, value: int, time_map: TimeMap, beat_snap: int
) -> song.LongNote:
    seconds = ticks_to_seconds(ticks)
    raw_beats = time_map.beats_at(seconds)
    beats = round_beats(raw_beats, beat_snap)
    eve_long = EveLong.from_value(value)
    seconds_duration = ticks_to_seconds(eve_long.duration)
    raw_beats_duration = time_map.beats_at(seconds + seconds_duration) - raw_beats
    beats_duration = round_beats(raw_beats_duration, beat_snap)
    position = song.NotePosition.from_index(eve_long.position)
    direction = VALUE_TO_DIRECTION[eve_long.direction]
    step_vector = song.TAIL_DIRECTION_TO_OUTWARDS_VECTOR[direction]
    raw_tail_pos = position + (eve_long.length * step_vector)
    tail_pos = song.NotePosition.from_raw_position(raw_tail_pos)
    return song.LongNote(
        time=beats, position=position, duration=beats_duration, tail_tip=tail_pos
    )


def make_hakus(
    hakus: List[int], end: int, time_map: TimeMap, beat_snap: int
) -> Optional[Set[song.BeatsTime]]:
    """Try to detect if the haku pattern is regular, in which case return None,
    otherwise return the parsed hakus"""
    roughly_rounded_hakus = make_raw_hakus(hakus, time_map, beat_snap=4)
    rough_end = beats_at_tick(end, time_map, beat_snap=4)
    if follows_regular_haku_pattern(roughly_rounded_hakus, rough_end):
        return None
    else:
        return make_raw_hakus(hakus, time_map, beat_snap)


def make_raw_hakus(
    hakus: List[int], time_map: TimeMap, beat_snap: int
) -> Set[song.BeatsTime]:
    return set(beats_at_tick(haku, time_map, beat_snap) for haku in hakus)


def follows_regular_haku_pattern(
    hakus: Set[song.BeatsTime], end_command: song.BeatsTime
) -> bool:
    """Regular hakus extend at least till the END command in a 4/4 rhythm"""
    if len(hakus) == 0:
        return False

    start = min(hakus)
    if (start % 1) != 0:
        return False

    haku_end = max(hakus)
    if (haku_end % 1) != 0:
        return False

    if haku_end < end_command:
        return False

    stop = haku_end + song.BeatsTime(1, 2)
    step = song.BeatsTime(1)
    regular = numeric_range(start, stop, step)
    return sorted(hakus) == list(regular)


def beats_at_tick(tick: int, time_map: TimeMap, beat_snap: int) -> song.BeatsTime:
    seconds = ticks_to_seconds(tick)
    raw_beats = time_map.beats_at(seconds)
    return round_beats(raw_beats, beat_snap)



from dataclasses import replace
from decimal import Decimal
from functools import partial, singledispatch
from pathlib import Path
from typing import Any, List, Set, Tuple, Union

from jubeatools import song as jbt
from jubeatools.utils import none_or

from ..tools import make_memon_folder_loader
from . import schema as memon


def _load_memon_1_0_0(raw_json: Any) -> jbt.Song:
    file: memon.File = memon.FILE_SCHEMA.load(raw_json)

    metadata = none_or(load_metadata, file.metadata) or jbt.Metadata()
    charts = {diff: load_chart(chart, file) for diff, chart in file.data.items()}
    if not file.timing:
        common_timing = None
        common_hakus = None
    else:
        timing = memon.Timing.fill_in_defaults(file.timing)
        common_timing = load_timing(timing)
        resolution = timing.resolution or 240
        load_hakus_with_res = partial(load_hakus, resolution=resolution)
        common_hakus = none_or(load_hakus_with_res, file.timing.hakus)

    return jbt.Song(
        metadata=metadata,
        charts=charts,
        common_timing=common_timing,
        common_hakus=common_hakus,
    )


load_memon_1_0_0 = make_memon_folder_loader(_load_memon_1_0_0)


def load_metadata(m: memon.Metadata) -> jbt.Metadata:
    result = jbt.Metadata(
        title=m.title,
        artist=m.artist,
        audio=none_or(Path, m.audio),
        cover=none_or(Path, m.jacket),
    )

    if m.preview is None:
        return result
    elif isinstance(m.preview, str):
        return replace(
            result,
            preview_file=Path(m.preview),
        )
    elif isinstance(m.preview, memon.PreviewSample):
        return replace(result, preview=jbt.Preview(m.preview.start, m.preview.duration))


def load_chart(c: memon.Chart, m: memon.File) -> jbt.Chart:
    applicable_timing = memon.Timing.fill_in_defaults(c.timing, m.timing)
    if not c.timing:
        timing = None
        hakus = None
    else:
        timing = load_timing(applicable_timing)
        resolution = applicable_timing.resolution or 240
        load_hakus_with_res = partial(load_hakus, resolution=resolution)
        hakus = none_or(load_hakus_with_res, c.timing.hakus)

    return jbt.Chart(
        level=c.level,
        timing=timing,
        hakus=hakus,
        notes=[load_note(n, c.resolution or 240) for n in c.notes],
    )


def load_hakus(h: List[memon.SymbolicTime], resolution: int) -> Set[jbt.BeatsTime]:
    return set(load_symbolic_time(t, resolution) for t in h)


def load_timing(t: memon.Timing) -> jbt.Timing:
    return jbt.Timing(
        events=load_bpms(t.bpms or [], t.resolution or 240),
        beat_zero_offset=t.offset or Decimal(0),
    )


def load_bpms(bpms: List[memon.BPMEvent], resolution: int) -> List[jbt.BPMEvent]:
    return [load_bpm(b, resolution) for b in bpms]


def load_bpm(bpm: memon.BPMEvent, resolution: int) -> jbt.BPMEvent:
    return jbt.BPMEvent(
        time=load_symbolic_time(bpm.beat, resolution),
        BPM=bpm.bpm,
    )


def load_symbolic_time(
    t: Union[int, Tuple[int, int, int]], resolution: int
) -> jbt.BeatsTime:
    if isinstance(t, int):
        return jbt.BeatsTime(t, resolution)
    else:
        return t[0] + jbt.BeatsTime(t[1], t[2])


@singledispatch
def load_note(note: memon.Note, resolution: int) -> Union[jbt.TapNote, jbt.LongNote]:
    raise ValueError(f"Unexpected argument : {note}")


@load_note.register
def load_tap_note(note: memon.TapNote, resolution: int) -> jbt.TapNote:
    return jbt.TapNote(
        time=load_symbolic_time(note.t, resolution),
        position=jbt.NotePosition.from_index(note.n),
    )


@load_note.register
def load_long_note(note: memon.LongNote, resolution: int) -> jbt.LongNote:
    position = jbt.NotePosition.from_index(note.n)
    return jbt.LongNote(
        time=load_symbolic_time(note.t, resolution),
        position=position,
        duration=load_symbolic_time(note.l, resolution),
        tail_tip=convert_6_notation_to_position(position, note.p),
    )


def convert_6_notation_to_position(pos: jbt.NotePosition, p: int) -> jbt.NotePosition:
    # horizontal
    if p < 3:
        if p < pos.x:
            x = p
        else:
            x = p + 1
        y = pos.y
    else:
        p -= 3
        x = pos.x
        if p < pos.y:
            y = p
        else:
            y = p + 1

    return jbt.NotePosition(x, y)



from importlib import resources
from pathlib import Path

from click.testing import CliRunner

from jubeatools import song as jbt
from jubeatools.formats import LOADERS, Format

from ..cli import convert
from . import data


def test_that_ommiting_beat_snap_works() -> None:
    """As pointed out by https://github.com/Stepland/jubeatools/issues/17"""
    runner = CliRunner()
    data_path = resources.files(data)
    with runner.isolated_filesystem(), resources.as_file(
        data_path / "Life Without You.eve"
    ) as p:
        result = runner.invoke(
            convert, [str(p.resolve(strict=True)), "out.txt", "-f", "memo2"]
        )
        if result.exception:
            raise result.exception
        assert result.exit_code == 0


def test_that_is_flag_works_the_way_intended() -> None:
    """It's unclear to me what the default value is for an option with
    is_flag=True"""
    data_path = resources.files(data)
    with resources.as_file(data_path / "Life Without You.eve") as p:
        called_with_the_flag = convert.make_context(
            "convert",
            [str(p.resolve(strict=True)), "out.txt", "-f", "memo2", "--circlefree"],
        )
        assert called_with_the_flag.params["dumper_options"]["circle_free"] is True

        called_without_the_flag = convert.make_context(
            "convert", [str(p.resolve(strict=True)), "out.txt", "-f", "memo2"]
        )
        dumper_options = called_without_the_flag.params.get("dumper_options")
        if dumper_options is not None:
            circle_free = dumper_options.get("circle_free")
            assert not circle_free


def test_that_the_merge_option_works_for_memon_files() -> None:
    runner = CliRunner()
    data_path = resources.files(data)
    with runner.isolated_filesystem(), resources.as_file(
        data_path / "memon_merge"
    ) as p:
        result = runner.invoke(
            convert,
            [
                "--input-format",
                "memon:v0.1.0",
                str(p.resolve(strict=True)),
                "--merge",
                "out.memon",
                "-f",
                "memon:v0.1.0",
            ],
        )
        if result.exception:
            raise result.exception
        assert result.exit_code == 0

        memon_loader = LOADERS[Format.MEMON_0_1_0]
        bsc = memon_loader(p / "Sky Bus For Hire BSC.memon")
        adv = memon_loader(p / "Sky Bus For Hire ADV.memon")
        ext = memon_loader(p / "Sky Bus For Hire EXT.memon")
        merged_by_cli = LOADERS[Format.MEMON_0_1_0](Path("out.memon"))
        merged_with_python = jbt.Song.from_monochart_instances(bsc, adv, ext)
        assert merged_by_cli == merged_with_python
